from __future__ import annotations

from dissect.vmfs.descriptor import DirEntry, FileDescriptor
from dissect.vmfs.exception import (
    Error,
    FileNotFoundError,
    InvalidHeader,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.vmfs.lvm import LVM, Device, Volume
from dissect.vmfs.vmfs import VMFS

__all__ = [
    "LVM",
    "VMFS",
    "Device",
    "DirEntry",
    "Error",
    "FileDescriptor",
    "FileNotFoundError",
    "InvalidHeader",
    "NotADirectoryError",
    "NotASymlinkError",
    "Volume",
]



from __future__ import annotations

from textwrap import dedent
from typing import TYPE_CHECKING, BinaryIO

from dissect.vmfs.address import (
    FileBlockAddr,
    FileDescriptorAddr,
    JournalBlockAddr,
    LargeFileBlockAddr,
    PointerBlock2Addr,
    PointerBlock2Addr64,
    PointerBlockAddr,
    PointerBlockAddr64,
    SmallFileBlockAddr,
    SubBlockAddr,
    SubBlockAddr64,
    address_type,
)
from dissect.vmfs.c_vmfs import FS3_Config, FS3_DescriptorType, FS3_ResourceTypeID, c_vmfs
from dissect.vmfs.util import bsf

if TYPE_CHECKING:
    from dissect.vmfs.vmfs import VMFS, FileDescriptor


class ResourceFile:
    """VMFS resource file implementation.

    Resource files of different types need different interpretation of the resource data. Some may not have
    any resource data at all, using only the bitmaps in the resource file to track allocation.

    Resource files start with a metadata header (``FS3_ResFileMetadata``) that contains information about the layout
    of the file, such as the number of resources, their size, and their organization into clusters and cluster groups.

    Following the metadata header, at an offset defined in ``clusterGroupOffset``, starts the first cluster group.
    Each cluster group contains a ``clustersPerGroup`` number of clusters. Each cluster group starts with an
    array of ``FS3_ResourceCluster`` structures, one for every cluster in the group. This structure contains a lock
    and a cluster metadata header, which has resource allocation information about the cluster.
    The resource data starts after the cluster metadata headers. To get to the correct resource data,
    you need to calculate the offset based on the cluster group, cluster and resource indices.

    Args:
        vmfs: The VMFS instance this resource file belongs to.
        resource_type: The type of the resource file.
        address: The file descriptor address of the resource file.
        fh: A file-like object to read the resource data from.
        fd: An optional file descriptor for the resource file, if available.
        metadata_offset: The offset in the file where the resource metadata starts. Defaults to 0.
    """

    def __init__(
        self,
        vmfs: VMFS,
        resource_type: FS3_ResourceTypeID,
        address: int,
        fh: BinaryIO,
        fd: FileDescriptor | None = None,
        metadata_offset: int = 0,
    ):
        self.vmfs = vmfs
        self.type = resource_type
        self.address = address
        self.fh = fh
        self.fd = fd

        self.fh.seek(metadata_offset)
        self.metadata = c_vmfs.FS3_ResFileMetadata(self.fh)
        if self.vmfs.is_vmfs6 and self.metadata.signature != c_vmfs.FS3_RFMD_SIGNATURE:
            raise ValueError("Invalid resource metadata signature")

        self._resource_per_cluster_shift = bsf(self.metadata.resourcesPerCluster)

        # Resource clusters contain an array of lock and metadata headers for each cluster in the group
        # Roughly looks like this:
        # struct FS3_ResourceCluster {
        #     FS3_DiskBlock clusterLock;
        #     FS3_DiskBlock clusterMeta;
        # };
        # Resource data starts after the cluster metadata headers.
        if self.vmfs.is_vmfs5:
            # VMFS5 uses a fixed metadata alignment of 512 bytes
            # We have two disk blocks per cluster (lock and metadata header), so the total header size is 1024 bytes
            self._cluster_header_size = 1024
        else:
            # VMFS6 uses a variable metadata alignment
            # Still have two disk blocks per cluster, so the total header size is 2 * mdAlignment
            self._cluster_header_size = 2 * self.vmfs.descriptor.mdAlignment

        # The offset from a cluster group start to the resource data is the combined size of all cluster headers
        self._cluster_data_offset = self.metadata.clustersPerGroup * self._cluster_header_size
        self._cluster_size = self.metadata.resourcesPerCluster * self.metadata.resourceSize

    def debug(self) -> str:
        """Return a debug string for this resource file.

        Mimicks ``vmkfstool -D`` output.
        """
        md = self.metadata
        fd = self.vmfs.file_descriptor(self.address)

        return (
            fd.debug()
            + "\n"
            + dedent(f"""
        {(md.numResourcesHi << 32) | md.numResourcesLo} resources, each of size {md.resourceSize}
        Organized as {md.numClusterGroups} CGs, {md.clustersPerClusterGroup} C/CG and {md.resourcesPerCluster} R/C
        CGsize {md.clusterGroupSize}. 0th CG at {md.firstClusterGroupOffset}
        """).strip()
        )

    def parse_address(self, address: int) -> tuple[int, int]:
        """Parse an address into a cluster/resource pair to use for looking up a resource."""
        raise NotImplementedError

    def resolve_address(self, address: int) -> int:
        """Resolve an address to a volume address."""
        offset_in_file = self._resource_offset_from_address(address)
        if not self.fd:
            raise ValueError("Need a file descriptor to resolve to a volume address")
        return self.fd._resolve_offset(offset_in_file)[0]

    def _cluster_group_offset(self, group: int) -> int:
        """Calculate the offset of a cluster group in the resource file."""
        md = self.metadata

        if self.vmfs.is_vmfs5 or (self.vmfs.is_vmfs6 and md.flags & 2 == 0):
            return (
                # Base offset for cluster groups
                md.clusterGroupOffset
                # Offset for the cluster group
                + (group * md.clusterGroupSize)
            )

        parent_resources_per_group = (md.parentClustersPerGroup * md.parentResourcesPerCluster) // md.clustersPerGroup
        parent_group, parent_cluster_in_group = divmod(group, parent_resources_per_group)

        return (
            # Base offset for cluster groups
            md.clusterGroupOffset
            # Offset for the parent cluster group
            + (parent_group * md.parentClusterGroupSize)
            # Skip the cluster headers for the parent cluster group
            + (md.parentClustersPerGroup * self._cluster_header_size)
            # Offset for the cluster group within the parent cluster group
            + (parent_cluster_in_group * md.clusterGroupSize)
        )

    def _cluster_header_offset(self, cluster: int) -> int:
        """Calculate the offset of a cluster header in the resource file."""
        md = self.metadata

        if self.vmfs.is_vmfs5 or (self.vmfs.is_vmfs6 and md.flags & 2 == 0):
            group, cluster_in_group = divmod(cluster, md.clustersPerGroup)
            return (
                # Base offset for cluster groups
                md.clusterGroupOffset
                # Offset for the cluster group
                + (group * md.clusterGroupSize)
                # Offset for the cluster header within the group
                + (cluster_in_group * self._cluster_header_size)
            )

        parent_resources_per_group = md.parentClustersPerGroup * md.parentResourcesPerCluster
        parent_group, parent_cluster_in_group = divmod(cluster, parent_resources_per_group)

        return (
            # Base offset for cluster groups
            md.clusterGroupOffset
            # Offset for the parent cluster group
            + (parent_group * md.parentClusterGroupSize)
            # Skip the cluster headers for the parent cluster group
            + (md.parentClustersPerGroup * self._cluster_header_size)
            # Offset to the cluster in the parent cluster group
            + parent_cluster_in_group
        )

    def _resource_offset(self, cluster: int, resource: int) -> int:
        """Calculate the offset of a specific resource into the resource file based on its address."""
        md = self.metadata
        group, cluster_in_group = divmod(cluster, md.clustersPerGroup)

        return (
            # Base offset for cluster groups
            md.clusterGroupOffset
            # Offset for the cluster group
            + (group * md.clusterGroupSize)
            # Offset for the cluster header within the group
            + (cluster_in_group * self._cluster_size)
            # Offset for the resource data of the cluster
            + self._cluster_data_offset
            # Offset for the specific resource within the cluster
            + (resource * md.resourceSize)
        )

    def _resource_offset_from_address(self, address: int) -> int:
        """Calculate the offset of a specific resource into the resource file based on its address."""
        cluster, resource = self.parse_address(address)
        return self._resource_offset(cluster, resource)

    def read(self, address: int) -> bytes:
        """Read and return the resource belonging to the given address."""
        # Prefer reading from the volume if we have it
        if self.vmfs.fh:
            offset = self.resolve_address(address)
            self.vmfs.fh.seek(offset)
            return self.vmfs.fh.read(self.metadata.resourceSize)

        # Otherwise, read from the resource file
        offset = self._resource_offset_from_address(address)
        self.fh.seek(offset)
        return self.fh.read(self.metadata.resourceSize)


class FileBlockResourceVMFS5(ResourceFile):
    """(Small) file block resource (``.fbb.sf``)."""

    def parse_address(self, address: int) -> tuple[int, int]:
        block = FileBlockAddr.parse(address)
        return divmod(block, self.metadata.resourcesPerCluster)

    def resolve_address(self, address: int) -> int:
        """Resolve a file block address to a volume address."""
        block = FileBlockAddr.parse(address)
        return block * self.fd.block_size


class SmallFileBlockResourceVMFS6(ResourceFile):
    """Small file block resource (``.fbb.sf``) for VMFS6."""

    def parse_address(self, address: int) -> tuple[int, int]:
        return SmallFileBlockAddr.parse(address)

    def resolve_address(self, address: int) -> int:
        """Resolve a small file block address to a volume address."""
        cluster, resource = self.parse_address(address)
        return ((cluster * self.metadata.resourcesPerCluster) + resource) << self.vmfs._file_block_size_shift


class SubBlockResourceVMFS5(ResourceFile):
    """Sub-block resource (``.sbc.sf``) for VMFS5."""

    def parse_address(self, address: int) -> tuple[int, int]:
        return SubBlockAddr.parse(address, dense=FS3_Config.DENSE_SBPC in self.vmfs.descriptor.config)


class SubBlockResourceVMFS6(ResourceFile):
    """Sub-block resource (``.sbc.sf``) for VMFS6.

    Uses 64-bit addressing.
    """

    def parse_address(self, address: int) -> tuple[int, int]:
        return SubBlockAddr64.parse(address)


class PointerBlockResourceVMFS5(ResourceFile):
    """Pointer block resource (``.pbc.sf``)."""

    def parse_address(self, address: int) -> tuple[int, int]:
        return PointerBlockAddr.parse(address)


class PointerBlockResourceVMFS6(ResourceFile):
    """Pointer block resource (``.pbc.sf``) for VMFS6.

    Uses 64-bit addressing.
    """

    def parse_address(self, address: int) -> tuple[int, int]:
        return PointerBlockAddr64.parse(address)


class FileDescriptorResourceVMFS5(ResourceFile):
    """File descriptor resource (``.fdc.sf``) for VMFS5."""

    def parse_address(self, address: int) -> tuple[int, int]:
        return FileDescriptorAddr.parse(address)


class FileDescriptorResourceVMFS6(ResourceFile):
    """File descriptor resource (``.fdc.sf``) for VMFS6."""

    def parse_address(self, address: int) -> tuple[int, int]:
        return FileDescriptorAddr.parse(address)


class PointerBlock2ResourceVMFS5(ResourceFile):
    """Pointer block 2 resource (``.pb2.sf``) for VMFS5."""

    def parse_address(self, address: int) -> tuple[int, int]:
        return PointerBlock2Addr.parse(address)


class PointerBlock2ResourceVMFS6(ResourceFile):
    """Pointer block 2 resource (``.pb2.sf``) for VMFS6.

    Uses 64-bit addressing.
    """

    def parse_address(self, address: int) -> tuple[int, int]:
        return PointerBlock2Addr64.parse(address)


class JournalBlockResourceVMFS6(ResourceFile):
    """Journal block resource (``.jbc.sf``)."""

    def parse_address(self, address: int) -> tuple[int, int]:
        return JournalBlockAddr.parse(address)


class LargeFileBlockResourceVMFS6(ResourceFile):
    """Large file block resource (``.fbb.sf``)."""

    def parse_address(self, address: int) -> tuple[int, int]:
        block = LargeFileBlockAddr.parse(address)
        cluster = block >> self._resource_per_cluster_shift
        resource = block & (self.metadata.resourcesPerCluster - 1)
        return cluster, resource

    def resolve_address(self, address: int) -> int:
        """Resolve a large file block address to a volume address."""
        cluster, resource = self.parse_address(address)
        block_shift = self.vmfs._file_block_size_shift + self.vmfs.descriptor.sfbToLfbShift
        return ((cluster << self._resource_per_cluster_shift) + resource) << block_shift


RESOURCE_TYPE_MAP_VMFS5 = {
    FS3_ResourceTypeID.FILE_BLOCK: FileBlockResourceVMFS5,
    FS3_ResourceTypeID.SUB_BLOCK: SubBlockResourceVMFS5,
    FS3_ResourceTypeID.PTR_BLOCK: PointerBlockResourceVMFS5,
    FS3_ResourceTypeID.FILE_DESC: FileDescriptorResourceVMFS5,
    FS3_ResourceTypeID.PTR2_BLOCK: PointerBlock2ResourceVMFS5,
}

RESOURCE_TYPE_MAP_VMFS6 = {
    FS3_ResourceTypeID.SMALL_FILE_BLOCK: SmallFileBlockResourceVMFS6,
    FS3_ResourceTypeID.SUB_BLOCK: SubBlockResourceVMFS6,
    FS3_ResourceTypeID.PTR_BLOCK: PointerBlockResourceVMFS6,
    FS3_ResourceTypeID.FILE_DESC: FileDescriptorResourceVMFS6,
    FS3_ResourceTypeID.PTR2_BLOCK: PointerBlock2ResourceVMFS6,
    FS3_ResourceTypeID.JOURNAL_BLOCK: JournalBlockResourceVMFS6,
    FS3_ResourceTypeID.LARGE_FILE_BLOCK: LargeFileBlockResourceVMFS6,
}


RESOURCE_TYPE_ABBREVIATIONS = {
    FS3_ResourceTypeID.FILE_BLOCK: "FB",
    FS3_ResourceTypeID.SMALL_FILE_BLOCK: "SFB",
    FS3_ResourceTypeID.SUB_BLOCK: "SB",
    FS3_ResourceTypeID.PTR_BLOCK: "PB",
    FS3_ResourceTypeID.FILE_DESC: "FD",
    FS3_ResourceTypeID.PTR2_BLOCK: "PB2",
    FS3_ResourceTypeID.JOURNAL_BLOCK: "JB",
    FS3_ResourceTypeID.LARGE_FILE_BLOCK: "LFB",
}


class ResourceManager:
    """Manager for VMFS resources.

    Utility class for keeping track of opened resource files and providing access to them.
    """

    def __init__(self, vmfs: VMFS):
        self.vmfs = vmfs
        self.resources = {}

    def __repr__(self) -> str:
        return (
            f"<ResourceManager {' '.join(RESOURCE_TYPE_ABBREVIATIONS[FS3_ResourceTypeID(r)] for r in self.resources)}>"
        )

    def __getitem__(self, type: FS3_ResourceTypeID | int) -> ResourceFile:
        """Get a resource file by its type."""
        resource_type = FS3_ResourceTypeID(type)
        if resource_type not in self.resources:
            raise KeyError(f"No resource opened for type {resource_type}")
        return self.resources[resource_type.value]

    def open(
        self,
        type: FS3_ResourceTypeID,
        address: int | None = None,
        fileobj: BinaryIO | None = None,
        metadata_offset: int = 0,
    ) -> None:
        """Open a resource file of the given type.

        Args:
            type: The type of the resource to open.
            address: The address of the resource to open. If not provided, `fileobj` must be provided.
            fileobj: A file-like object to read the resource from. If not provided, `address` must be provided.
            metadata_offset: The offset in the file where the resource metadata starts. Defaults to 0.
                This is used for child metadata that share the same resource file.
        """
        if type.value in self.resources:
            raise ValueError(f"Resource already opened: {type}")

        resource_type_map = RESOURCE_TYPE_MAP_VMFS5 if self.vmfs.is_vmfs5 else RESOURCE_TYPE_MAP_VMFS6
        if type not in resource_type_map:
            raise TypeError(f"Don't know how to open resource: {type}")

        if not address and not fileobj:
            raise ValueError(f"No address or file object for resource: {type}")

        fd = None
        if not fileobj:
            # Open the system file descriptor
            fd = self.vmfs._get_sfd(address)
            # Do a sanity check on the file descriptor
            if (
                fd.metadata.descAddr != address
                or fd.block_size != self.vmfs.file_block_size
                or fd.type != FS3_DescriptorType.SYSFILE
            ):
                raise ValueError(f"Invalid file descriptor for resource {type}: {fd}")

            # Fix up metadata for older VMFS versions
            # Untested, but just in case
            if self.vmfs.major_version <= 19:
                fd.metadata.blockOffsetShift = bsf(fd.metadata.blockSize)
                fd.metadata.tbzGranularityShift = fd.metadata.blockOffsetShift

            fileobj = fd.open()

        self.resources[type.value] = resource_type_map[type](
            self.vmfs,
            type,
            address,
            fileobj,
            fd=fd,
            metadata_offset=metadata_offset,
        )

    def close(self, type: FS3_ResourceTypeID) -> None:
        """Close a resource file of the given type.

        Args:
            type: The type of the resource to close.
        """
        if type.value not in self.resources:
            raise ValueError(f"No resource opened for type {type}")

        del self.resources[type.value]

    def has(self, type: FS3_ResourceTypeID) -> bool:
        """Check if a resource of the given type is opened.

        Args:
            type: The type of the resource to check.
        """
        return type.value in self.resources

    def get(self, type: FS3_ResourceTypeID | int) -> ResourceFile | None:
        """Get a resource file by its type.

        Args:
            type: The type of the resource to get.
        """
        type = FS3_ResourceTypeID(type)
        return self.resources.get(type.value)

    def read(self, address: int) -> bytes:
        """Read a resource from the given address, automatically selecting the correct resource from the address.

        Args:
            address: The address of the resource to read.
        """
        return self.get(address_type(address)).read(address)

    @property
    def FB(self) -> FileBlockResourceVMFS5 | SmallFileBlockResourceVMFS6:
        """Get the file block resource (``.fbb.sf``, VMFS5)."""
        return self.get(FS3_ResourceTypeID.FILE_BLOCK)

    @property
    def SFB(self) -> FileBlockResourceVMFS5 | SmallFileBlockResourceVMFS6:
        """Get the small file block resource (``.fbb.sf``, VMFS6)."""
        return self.get(FS3_ResourceTypeID.SMALL_FILE_BLOCK)

    @property
    def SB(self) -> SubBlockResourceVMFS5 | SubBlockResourceVMFS6:
        """Get the sub-block resource (``.sbc.sf``)."""
        return self.get(FS3_ResourceTypeID.SUB_BLOCK)

    @property
    def PB(self) -> PointerBlockResourceVMFS5 | PointerBlockResourceVMFS6:
        """Get the pointer block resource (``.pbc.sf``)."""
        return self.get(FS3_ResourceTypeID.PTR_BLOCK)

    @property
    def FD(self) -> FileDescriptorResourceVMFS5 | FileDescriptorResourceVMFS6:
        """Get the file descriptor resource (``.fdc.sf``)."""
        return self.get(FS3_ResourceTypeID.FILE_DESC)

    @property
    def PB2(self) -> PointerBlock2ResourceVMFS5 | PointerBlock2ResourceVMFS6:
        """Get the pointer block 2 resource (``.pb2.sf``)."""
        return self.get(FS3_ResourceTypeID.PTR2_BLOCK)

    @property
    def JB(self) -> JournalBlockResourceVMFS6:
        """Get the journal block resource (``.jbc.sf``)."""
        return self.get(FS3_ResourceTypeID.JOURNAL_BLOCK)

    @property
    def LFB(self) -> LargeFileBlockResourceVMFS6:
        """Get the large file block resource (``.fbb.sf``)."""
        return self.get(FS3_ResourceTypeID.LARGE_FILE_BLOCK)



# Generated by cstruct-stubgen
from typing import BinaryIO, Literal, TypeAlias, overload

import dissect.cstruct as __cs__

class _c_lvm(__cs__.cstruct):
    DISK_BLOCK_SIZE_512B: Literal[512] = ...
    FS_PLIST_DEF_MAX_PARTITIONS: Literal[32] = ...
    LVM_MAGIC_NUMBER: Literal[3221344269] = ...
    LVM_INITIAL_MAJOR_VERSION: Literal[3] = ...
    LVM_MAJOR_VERSION_ESX60: Literal[6] = ...
    LVM_MAJOR_VERSION_ESX50: Literal[5] = ...
    LVM_IS_LVM6: Literal["((majorVersion == LVM_MAJOR_VERSION_ESX60) ? TRUE : FALSE)"] = ...
    LVM_DEV_HEADER_OFFSET: Literal[1048576] = ...
    LVM_MD_ALIGNMENT_DEFAULT: Literal[4096] = ...
    LVM_MAX_VOLUME_LABEL_LENGTH: Literal[64] = ...
    LVM_SIZEOF_LVM_DEVMETA_LVM5: Literal[512] = ...
    LVM_SIZEOF_LVM_DEVMETA_LVM6: Literal["(mdAlignment)"] = ...
    LVM_SIZEOF_LVM_DEVMETA: Literal[
        "(LVM_IS_LVM6(majorVersion) ? LVM_SIZEOF_LVM_DEVMETA_LVM6(mdAlignment) : LVM_SIZEOF_LVM_DEVMETA_LVM5)"
    ] = ...
    LVM_SIZEOF_EXTVOLMETA: Literal[256] = ...
    LVM_SIZEOF_VTENTRY: Literal[512] = ...
    LVM_SIZEOF_PTENTRY: Literal[128] = ...
    LVM_SIZEOF_SDTENTRY: Literal[256] = ...
    LVM_MAX_VOLUMES_PER_DEV_LVM5: Literal[512] = ...
    LVM_MAX_VOLUMES_PER_DEV_LVM6: Literal[1] = ...
    LVM_MAX_VOLUMES_PER_DEV: Literal[
        "(LVM_IS_LVM6(majorVersion) ? LVM_MAX_VOLUMES_PER_DEV_LVM6 : LVM_MAX_VOLUMES_PER_DEV_LVM5)"
    ] = ...
    LVM_UNUSED_MD_SECTORS_LVM5: Literal[512] = ...
    LVM_UNUSED_MD_SECTORS_LVM6: Literal[1023] = ...
    LVM_UNUSED_MD_SIZE_LVM5: Literal[262144] = ...
    LVM_UNUSED_MD_SIZE_LVM6: Literal[523776] = ...
    LVM_RESERVED_SIZE_LVM5: Literal[253952] = ...
    LVM_RESERVED_SIZE_LVM6: Literal[515584] = ...
    LVM_RESERVED_SIZE: Literal["(LVM_IS_LVM6(majorVersion) ? LVM_RESERVED_SIZE_LVM6 : LVM_RESERVED_SIZE_LVM5)"] = ...
    LVM_PES_PER_BITMAP: Literal[8192] = ...
    LVM_PE_BITMAP_SIZE_LVM5: Literal[1024] = ...
    LVM_PE_BITMAP_SIZE_LVM6: Literal["(MAX(mdAlignment, LVM_PE_BITMAP_SIZE_LVM5))"] = ...
    LVM_PE_BITMAP_SIZE: Literal[
        "(LVM_IS_LVM6(majorVersion) ? LVM_PE_BITMAP_SIZE_LVM6(mdAlignment) : LVM_PE_BITMAP_SIZE_LVM5)"
    ] = ...
    Bool: TypeAlias = _c_lvm.uint8
    class LVM_VolState(__cs__.Enum):
        LVM_INVALID = ...
        LVM_NORMAL = ...
        LVM_RESIG_PENDING = ...
        LVM_RESIG_DONE = ...
        LVM_SENTINEL = ...

    class UUID(__cs__.Structure):
        timeLo: _c_lvm.uint32
        timeHi: _c_lvm.uint32
        rand: _c_lvm.uint16
        macAddr: __cs__.CharArray
        @overload
        def __init__(
            self,
            timeLo: _c_lvm.uint32 | None = ...,
            timeHi: _c_lvm.uint32 | None = ...,
            rand: _c_lvm.uint16 | None = ...,
            macAddr: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class SCSI_DiskId(__cs__.Structure):
        type: _c_lvm.uint8
        len: _c_lvm.uint8
        lun: _c_lvm.uint16
        deviceType: _c_lvm.uint8
        scsiLevel: _c_lvm.uint8
        vendor: __cs__.CharArray
        model: __cs__.CharArray
        revision: __cs__.CharArray
        id: __cs__.CharArray
        @overload
        def __init__(
            self,
            type: _c_lvm.uint8 | None = ...,
            len: _c_lvm.uint8 | None = ...,
            lun: _c_lvm.uint16 | None = ...,
            deviceType: _c_lvm.uint8 | None = ...,
            scsiLevel: _c_lvm.uint8 | None = ...,
            vendor: __cs__.CharArray | None = ...,
            model: __cs__.CharArray | None = ...,
            revision: __cs__.CharArray | None = ...,
            id: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class LVM_DevMetadata(__cs__.Structure):
        magic: _c_lvm.uint32
        majorVersion: _c_lvm.uint32
        minorVersion: _c_lvm.uint32
        diskID: _c_lvm.SCSI_DiskId
        diskBlockSize: _c_lvm.uint32
        totalBytes: _c_lvm.uint64
        numVolumes: _c_lvm.uint32
        numPEs: _c_lvm.uint32
        lastPEIndex: _c_lvm.uint32
        generation: _c_lvm.uint64
        dataOffset: _c_lvm.uint64
        devID: _c_lvm.UUID
        initTimeUS: _c_lvm.uint64
        modTimeUS: _c_lvm.uint64
        lockedBy: _c_lvm.UUID
        lockedWhenUS: _c_lvm.uint64
        _unknown0: _c_lvm.uint32
        numPEMaps: _c_lvm.uint32
        extDevMetadataOffset: _c_lvm.uint64
        mdAlignment: _c_lvm.uint32
        numPEs6: _c_lvm.uint32
        flags: _c_lvm.uint32
        @overload
        def __init__(
            self,
            magic: _c_lvm.uint32 | None = ...,
            majorVersion: _c_lvm.uint32 | None = ...,
            minorVersion: _c_lvm.uint32 | None = ...,
            diskID: _c_lvm.SCSI_DiskId | None = ...,
            diskBlockSize: _c_lvm.uint32 | None = ...,
            totalBytes: _c_lvm.uint64 | None = ...,
            numVolumes: _c_lvm.uint32 | None = ...,
            numPEs: _c_lvm.uint32 | None = ...,
            lastPEIndex: _c_lvm.uint32 | None = ...,
            generation: _c_lvm.uint64 | None = ...,
            dataOffset: _c_lvm.uint64 | None = ...,
            devID: _c_lvm.UUID | None = ...,
            initTimeUS: _c_lvm.uint64 | None = ...,
            modTimeUS: _c_lvm.uint64 | None = ...,
            lockedBy: _c_lvm.UUID | None = ...,
            lockedWhenUS: _c_lvm.uint64 | None = ...,
            _unknown0: _c_lvm.uint32 | None = ...,
            numPEMaps: _c_lvm.uint32 | None = ...,
            extDevMetadataOffset: _c_lvm.uint64 | None = ...,
            mdAlignment: _c_lvm.uint32 | None = ...,
            numPEs6: _c_lvm.uint32 | None = ...,
            flags: _c_lvm.uint32 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class LVM_ExtDevMetadata(__cs__.Structure):
        magic: _c_lvm.uint32
        numPEMaps: _c_lvm.uint32
        dataOffset: _c_lvm.uint64
        nextOffset: _c_lvm.uint64
        @overload
        def __init__(
            self,
            magic: _c_lvm.uint32 | None = ...,
            numPEMaps: _c_lvm.uint32 | None = ...,
            dataOffset: _c_lvm.uint64 | None = ...,
            nextOffset: _c_lvm.uint64 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class LVM_VolID(__cs__.Structure):
        uuid: _c_lvm.UUID
        snapID: _c_lvm.uint32
        @overload
        def __init__(self, uuid: _c_lvm.UUID | None = ..., snapID: _c_lvm.uint32 | None = ...): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class LVM_VolMetadata(__cs__.Structure):
        logicalSize: _c_lvm.uint64
        generation: _c_lvm.uint64
        state: _c_lvm.LVM_VolState
        name: __cs__.CharArray
        lvID: _c_lvm.LVM_VolID
        creationTimeUS: _c_lvm.uint64
        @overload
        def __init__(
            self,
            logicalSize: _c_lvm.uint64 | None = ...,
            generation: _c_lvm.uint64 | None = ...,
            state: _c_lvm.LVM_VolState | None = ...,
            name: __cs__.CharArray | None = ...,
            lvID: _c_lvm.LVM_VolID | None = ...,
            creationTimeUS: _c_lvm.uint64 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class LVM_ExtVolMetadata(__cs__.Structure):
        numDevs: _c_lvm.uint32
        pad: __cs__.CharArray
        consumedPEs: _c_lvm.uint32
        @overload
        def __init__(
            self,
            numDevs: _c_lvm.uint32 | None = ...,
            pad: __cs__.CharArray | None = ...,
            consumedPEs: _c_lvm.uint32 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class LVM_VolDescriptor(__cs__.Structure):
        volMeta: _c_lvm.LVM_VolMetadata
        volumeID: _c_lvm.uint32
        numPEs: _c_lvm.uint32
        firstPE: _c_lvm.uint64
        lastPE: _c_lvm.uint64
        modTimeUS: _c_lvm.uint64
        extVolMeta: _c_lvm.LVM_ExtVolMetadata
        pad: __cs__.CharArray
        @overload
        def __init__(
            self,
            volMeta: _c_lvm.LVM_VolMetadata | None = ...,
            volumeID: _c_lvm.uint32 | None = ...,
            numPEs: _c_lvm.uint32 | None = ...,
            firstPE: _c_lvm.uint64 | None = ...,
            lastPE: _c_lvm.uint64 | None = ...,
            modTimeUS: _c_lvm.uint64 | None = ...,
            extVolMeta: _c_lvm.LVM_ExtVolMetadata | None = ...,
            pad: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class LVM_VolTableEntry(__cs__.Structure):
        volDesc: _c_lvm.LVM_VolDescriptor
        @overload
        def __init__(self, volDesc: _c_lvm.LVM_VolDescriptor | None = ...): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class LVM_PEDescriptor(__cs__.Structure):
        peID: _c_lvm.uint32
        volumeID: _c_lvm.uint32
        pOffset: _c_lvm.uint64
        lOffset: _c_lvm.uint64
        length: _c_lvm.uint64
        version: _c_lvm.uint32
        @overload
        def __init__(
            self,
            peID: _c_lvm.uint32 | None = ...,
            volumeID: _c_lvm.uint32 | None = ...,
            pOffset: _c_lvm.uint64 | None = ...,
            lOffset: _c_lvm.uint64 | None = ...,
            length: _c_lvm.uint64 | None = ...,
            version: _c_lvm.uint32 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class LVM_PETableEntry(__cs__.Structure):
        used: _c_lvm.uint8
        peDesc: _c_lvm.LVM_PEDescriptor
        @overload
        def __init__(self, used: _c_lvm.uint8 | None = ..., peDesc: _c_lvm.LVM_PEDescriptor | None = ...): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class LVM_SDTableEntry(__cs__.Structure):
        deviceName: __cs__.CharArray
        @overload
        def __init__(self, deviceName: __cs__.CharArray | None = ...): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

# Technically `c_lvm` is an instance of `_c_lvm`, but then we can't use it in type hints
c_lvm: TypeAlias = _c_lvm



# References:
# - /usr/lib/vmware/vmkmod/lvmdriver
from __future__ import annotations

from bisect import bisect_right
from functools import cache
from operator import itemgetter
from typing import TYPE_CHECKING, BinaryIO

from dissect.util.stream import AlignedStream
from dissect.util.ts import from_unix_us

from dissect.vmfs.c_lvm import c_lvm
from dissect.vmfs.exception import InvalidHeader, VolumeNotAvailableError
from dissect.vmfs.util import vmfs_uuid

if TYPE_CHECKING:
    from collections.abc import Iterator


class LVM:
    """VMFS LVM implementation, supports LVM5 and LVM6.

    VMFS LVM is a logical volume manager that allows multiple physical devices to be combined into a single
    logical volume. Technically LVM supports multiple logical volumes, in fact LVM3 started with supporting 1024, later
    versions decreased it to 512. LVM6 only allows 1. In practice only one logical volume is ever used.

    Provide this class with file-like objects for all devices that make up the LVM, then access the
    :attr:`~LVM.volumes` attribute to get a list of logical volumes. A :class:`Volume` can be opened for reading
    by calling :meth:`Volume.open`.

    Args:
        fh: A file-like object or a list of file-like objects that constitute an LVM.
    """

    def __init__(self, fh: BinaryIO | Device | list[BinaryIO] | list[Device]):
        fhs = [fh] if type(fh) is not list else fh

        self.devices: list[Device] = []
        """List of :class:`Device` objects that make up the LVM."""
        self.volumes: list[Volume] = []
        """List of :class:`Volume` objects that are in the LVM."""

        for fh in fhs:
            device = Device(fh) if not isinstance(fh, Device) else fh
            self.devices.append(device)

        volume_map = {}
        for device in self.devices:
            for vol in device.volumes:
                lv_id = (vmfs_uuid(vol.volMeta.lvID.uuid), vol.volMeta.lvID.snapID)
                volume_map.setdefault(lv_id, []).append(device)

        self.volumes = [Volume(lv_uuid, lv_snap_id, devices) for (lv_uuid, lv_snap_id), devices in volume_map.items()]

    def __repr__(self) -> str:
        return f"<LVM devices={len(self.devices)} volumes={len(self.volumes)}>"


class Device:
    """VMFS LVM device implementation.

    Represents a single device in the LVM.

    LVM devices contain metadata that describes itself, the logical volumes and physical extents it contains.

    The metadata roughly looks like the following pseudo-structure:

    .. code-block:: c

        struct LVM_DeviceHeader {
            LVM_DevMetadata     devMeta;
            LVM_VolTableEntry   volTable[LVM_MAX_VOLUMES_PER_DEV];
            char                reserved[LVM_RESERVED_SIZE];
            LVM_SDTableEntry    sdTable[FS_PLIST_DEF_MAX_PARTITIONS];
            uint8               peBitmap[LVM_PE_BITMAP_SIZE];
            LVM_PETableEntry    peTable[LVM_PES_PER_BITMAP];
        };

    On versions prior to LVM6, it looks like the structure sizes are largely respected when calculating offsets to other
    tables. However, since LVM6 a specific field in ``LVM_DevMetadata`` is often used for this calculation.
    Because the real name of this field is unknown, we have decided to call it ``mdAlignment`` within this project,
    since it appears to be used in a similar way as in VMFS.

    The device metadata (``devMeta``) starts at a fixed offset (``0x00100000``), but since LVM6 may
    reference extended metadata at other offsets. The volume table (``volTable``) starts after ``devMeta``, which is
    ``0x00100000 + LVM_SIZEOF_LVM_DEVMETA``, where ``LVM_SIZEOF_LVM_DEVMETA`` is either ``512`` or since LVM6 the value
    of the ``mdAlignment`` field in the ``LVM_DevMetadata`` structure. Since LVM5 (I could not find evidence that LVM4
    exists) there exists a ``sdTable``, which is a table of device names, that starts at the end of the volume table.
    The ``peBitmap`` is a bitmap that describes which entries in the ``peTable`` are used, and starts at the end of the
    ``volTable``/``sdTable``. The ``peTable`` is a table of physical extents, which starts at the end of
    the ``peBitmap``. A pair of ``peBitmap`` and ``peTable`` repeats for ``numPEs`` times.

    The device metadata ``LVM_DevMetadata`` contains information about the device, including some identifiers and
    number of volumes and physical extents. There are also timestamps when the device was created and last modified.

    The volume descriptor ``LVM_VolDescriptor`` contains information about the logical volume, specific to that device.
    The ``LVM_VolMetadata`` structure contains metadata that is shared across all devices in the volume, but other
    fields in the descriptor are specific to that device (such as the first and last physical extent on that device).

    The "SD table" (storage device? SCSI disk?) is a table of device names that are part of the volume, which is only
    present in LVM5 and later, and only on the first device in the LVM (internally referred to as "devZero").

    The physical extent descriptors (``LVM_PEDescriptor``) contain information about the physical extents on the device,
    including the logical offset, physical offset and length of the extent, as well as a reference to the volume it
    belongs to. A device can have multiple physical extents, and the logical volume is constructed from these
    physical extents across all devices in the LVM.

    There can only be a maximum of 8 physical extent map/table pairs per metadata region (so 8 maps and 64k table
    entries). If more are needed, the device metadata will reference extended metadata regions, which are similarly
    laid out, but with a different offset. The extended metadata regions are linked together by the ``nextOffset`` field
    in the ``LVM_ExtDevMetadata`` structure.

    Args:
        fh: A file-like object of a LVM device.
    """

    def __init__(self, fh: BinaryIO):
        self.fh = fh

        self.fh.seek(c_lvm.LVM_DEV_HEADER_OFFSET)
        self.metadata = c_lvm.LVM_DevMetadata(self.fh)
        self.ext_metadata = []

        if self.metadata.magic != c_lvm.LVM_MAGIC_NUMBER:
            raise InvalidHeader(
                f"Invalid device header. Expected 0x{c_lvm.LVM_MAGIC_NUMBER:08x}, got 0x{self.metadata.magic:08x}"
            )

        ext_dev_metadata_offset = self.metadata.extDevMetadataOffset
        while ext_dev_metadata_offset != 0:
            self.fh.seek(ext_dev_metadata_offset)
            ext_meta = c_lvm.LVM_ExtDevMetadata(self.fh)

            if ext_meta.magic != c_lvm.LVM_MAGIC_NUMBER:
                raise InvalidHeader(
                    "Invalid extended metadata header. "
                    f"Expected 0x{c_lvm.LVM_MAGIC_NUMBER:08x}, got 0x{ext_meta.magic:08x}"
                )

            self.ext_metadata.append(ext_meta)
            ext_dev_metadata_offset = ext_meta.nextOffset

        self.major_version = self.metadata.majorVersion
        self.minor_version = self.metadata.minorVersion
        self.uuid = vmfs_uuid(self.metadata.devID)
        self.size = self.metadata.totalBytes

        if self.major_version < 6:
            self._device_metadata_size = c_lvm.LVM_SIZEOF_LVM_DEVMETA_LVM5
            self._max_volumes_per_device = c_lvm.LVM_MAX_VOLUMES_PER_DEV_LVM5
            self._pe_bitmap_size = c_lvm.LVM_PE_BITMAP_SIZE_LVM5
        else:
            # LVM_SIZEOF_LVM_DEVMETA_LVM6(mdAlignment) (mdAlignment)
            self._device_metadata_size = self.metadata.mdAlignment
            self._max_volumes_per_device = c_lvm.LVM_MAX_VOLUMES_PER_DEV_LVM6
            # LVM_PE_BITMAP_SIZE_LVM6(mdAlignment)  (MAX(mdAlignment, LVM_PE_BITMAP_SIZE_LVM5))
            self._pe_bitmap_size = max(self.metadata.mdAlignment, c_lvm.LVM_PE_BITMAP_SIZE_LVM5)

        # LVM_UNUSED_MD_SECTORS (1024 - (LVM_MAX_VOLUMES_PER_DEV))
        self._unused_md_sectors = 1024 - self._max_volumes_per_device
        # LVM_UNUSED_MD_SIZE    LVM_UNUSED_MD_SECTORS * DISK_BLOCK_SIZE_512B
        self._unused_md_size = self._unused_md_sectors * c_lvm.DISK_BLOCK_SIZE_512B
        # LVM_RESERVED_SIZE     (LVM_UNUSED_MD_SIZE - LVM_SIZEOF_SDTENTRY * FS_PLIST_DEF_MAX_PARTITIONS)
        self._reserved_size = self._unused_md_size - (c_lvm.LVM_SIZEOF_SDTENTRY * c_lvm.FS_PLIST_DEF_MAX_PARTITIONS)

        # Basically offsetof(LVM_DeviceHeader, volTable)
        self._offset_to_volume_table = self._device_metadata_size
        # Basically offsetof(LVM_DeviceHeader, sdTable)
        self._offset_to_sd_table = (
            self._offset_to_volume_table
            # Size of volume table
            + (self._max_volumes_per_device * c_lvm.LVM_SIZEOF_VTENTRY)
            # Size of reserved space
            + self._reserved_size
        )
        # Basically offsetof(LVM_DeviceHeader, peBitmap)
        self._offset_to_pe_bitmap = (
            self._offset_to_sd_table
            # Size of SD table
            + (c_lvm.LVM_SIZEOF_SDTENTRY * c_lvm.FS_PLIST_DEF_MAX_PARTITIONS)
        )

        if self.metadata.numVolumes > self._max_volumes_per_device:
            raise ValueError(f"Unsupported number of volumes in LVM metadata: {self.metadata.numVolumes}")

        # Basically offsetof(LVM_DeviceHeader, volTable)
        # LVM_DEV_HEADER_OFFSET + LVM_SIZEOF_LVM_DEVMETA(majorVersion,mdAlignment)
        volume_table_offset = c_lvm.LVM_DEV_HEADER_OFFSET + self._device_metadata_size

        self.fh.seek(volume_table_offset)
        self.volumes = [entry.volDesc for entry in c_lvm.LVM_VolTableEntry[self.metadata.numVolumes](self.fh)]

    def __repr__(self) -> str:
        return f"<Device uuid={self.uuid} size={self.metadata.totalBytes}>"

    def _iter_pe_offsets(self) -> Iterator[tuple[int, int]]:
        """Iterate over the offsets of the physical extent bitmaps and data offsets."""
        num_pe_maps = self.metadata.numPEMaps or 1

        # Basically offsetof(LVM_DeviceHeader, peBitmap)
        pe_bitmap_offset = c_lvm.LVM_DEV_HEADER_OFFSET + self._offset_to_pe_bitmap
        pe_data_offset = self.metadata.dataOffset

        ext_dev_metadata_offset = self.metadata.extDevMetadataOffset
        ext_meta_it = iter(self.ext_metadata)

        pe_table_size = c_lvm.LVM_PES_PER_BITMAP * c_lvm.LVM_SIZEOF_PTENTRY

        while True:
            for i in range(num_pe_maps):
                map_offset = pe_bitmap_offset + (i * (self._pe_bitmap_size + pe_table_size))
                data_offset = pe_data_offset + ((i << 13) << 28)

                yield map_offset, data_offset

            if (ext_meta := next(ext_meta_it, None)) is None:
                break

            num_pe_maps = ext_meta.numPEMaps
            pe_bitmap_offset = ext_dev_metadata_offset + self._offset_to_pe_bitmap
            pe_data_offset = ext_meta.dataOffset

            ext_dev_metadata_offset = ext_meta.nextOffset

    def _iter_pe(self) -> Iterator[c_lvm.LVM_PETableEntry]:
        """Iterate over the physical extent entries in the device."""
        pe_idx = 0

        for map_offset, _ in self._iter_pe_offsets():
            table_offset = map_offset + self._pe_bitmap_size
            for _ in range(min(self.metadata.numPEs - pe_idx, c_lvm.LVM_PES_PER_BITMAP)):
                self.fh.seek(table_offset)
                yield c_lvm.LVM_PETableEntry(self.fh)
                table_offset += c_lvm.LVM_SIZEOF_PTENTRY
                pe_idx += 1

            if pe_idx == self.metadata.numPEs:
                break


class Volume:
    """Logical volume in a VMFS LVM.

    Represents a logical volume that is constructed from one or more devices.

    Args:
        uuid: The UUID of the volume.
        snap_id: The snapshot ID of the volume.
        devices: A list of :class:`Device` objects that make up the volume. Must contain at least one device.
    """

    def __init__(self, uuid: str, snap_id: int, devices: list[Device]):
        self.uuid = uuid
        self.snap_id = snap_id
        self.devices = devices

        if not devices:
            raise ValueError("Need at least one device to construct a volume")

        self._volume_descriptor_map = {}
        for device in devices:
            for volume in device.volumes:
                if (vmfs_uuid(volume.volMeta.lvID.uuid), volume.volMeta.lvID.snapID) == (uuid, snap_id):
                    self._volume_descriptor_map[device.uuid] = volume

        self.devices.sort(key=lambda d: self._volume_descriptor_map[d.uuid].firstPE)

        dev_zero = self.devices[0]
        vol_desc = self._volume_descriptor_map[dev_zero.uuid]
        self.size = vol_desc.volMeta.logicalSize
        self.generation = vol_desc.volMeta.generation
        self.state = vol_desc.volMeta.state
        self.name = vol_desc.volMeta.name.split(b"\x00", 1)[0].decode()
        self.creation_ts = from_unix_us(vol_desc.volMeta.creationTimeUS)

        # Valid devZero
        if dev_zero.major_version >= 5 and vol_desc.firstPE == 0:
            dev_zero.fh.seek(c_lvm.LVM_DEV_HEADER_OFFSET + dev_zero._offset_to_sd_table)

            self.device_names = [
                c_lvm.LVM_SDTableEntry(dev_zero.fh).deviceName.split(b"\x00", 1)[0].decode()
                for _ in range(vol_desc.extVolMeta.numDevs)
            ]
        else:
            self.device_names = []

        self.dataruns = cache(self.dataruns)

    def __repr__(self) -> str:
        return f"<Volume name={self.name!r} size={self.size} state={self.state.name}>"

    def is_valid(self) -> bool:
        """Check if the volume is valid and can be opened for reading."""
        # Should have at least one device
        if not self.devices:
            return False

        vol_desc = self._volume_descriptor_map.get(self.devices[0].uuid)

        # Should have devZero
        if vol_desc.firstPE != 0:
            return False

        # Should have the expected number of devices
        if vol_desc.extVolMeta.numDevs != len(self.devices):
            return False

        # Must have a continuous range of physical extents
        current_pe = 0
        for device in self.devices:
            if self._volume_descriptor_map[device.uuid].firstPE != current_pe:
                return False

            current_pe = self._volume_descriptor_map[device.uuid].lastPE + 1

        return True

    def _iter_pe(self) -> Iterator[tuple[int, int, int, Device]]:
        """Iterate over the physical extents that belong to this volume."""

        def _iter_pe_raw() -> Iterator[tuple[int, int, int, Device]]:
            for device in self.devices:
                vol_desc = self._volume_descriptor_map[device.uuid]

                for pe in device._iter_pe():
                    if not pe.used:
                        continue

                    pe_desc = pe.peDesc
                    if pe_desc.volumeID != vol_desc.volumeID:
                        continue

                    yield pe_desc.lOffset, pe_desc.pOffset, pe_desc.length, device

        yield from sorted(_iter_pe_raw(), key=itemgetter(0))

    def dataruns(self) -> list[tuple[int, int, int, Device]]:
        """Get the dataruns of the volume.

        Returns:
            A list of tuples of ``(logical_offset, physical_offset, size, device)`` for each run of contiguous data.
        """
        runs = []
        run_logical_offset = None
        run_physical_offset = None
        run_size = None
        run_device = None

        expected_offset = 0
        for logical_offset, physical_offset, length, device in self._iter_pe():
            if logical_offset != expected_offset:
                raise ValueError(
                    f"Found hole in volume physical extents: expected {expected_offset:#x}, got {logical_offset:#x}"
                )

            expected_offset += length

            if run_logical_offset is None:
                run_logical_offset = logical_offset
                run_physical_offset = physical_offset
                run_size = length
                run_device = device
                continue

            if run_physical_offset + run_size == physical_offset and run_device == device:
                # Extend the current run
                run_size += length
            else:
                # Save the current run and start a new one
                runs.append((run_logical_offset, run_physical_offset, run_size, run_device))
                run_logical_offset = logical_offset
                run_physical_offset = physical_offset
                run_size = length
                run_device = device

        if run_logical_offset is not None:
            runs.append((run_logical_offset, run_physical_offset, run_size, run_device))

        return runs

    def open(self) -> VolumeStream:
        """Open a read-only stream for the volume."""
        if not self.is_valid():
            raise VolumeNotAvailableError("Volume is in an invalid state and cannot be opened for reading")
        return VolumeStream(self)


class VolumeStream(AlignedStream):
    """Read-only stream that allows reading from VMFS LVM volumes.

    Args:
        volume: The :class:`Volume` to provide a stream for.
    """

    def __init__(self, volume: Volume):
        self.volume = volume
        self.runs = volume.dataruns()
        self._lookup = [logical_offset for logical_offset, _, _, _ in self.runs if logical_offset != 0]
        super().__init__(volume.size)

    def _read(self, offset: int, length: int) -> bytes:
        result = []

        run_idx = bisect_right(self._lookup, offset)
        while length > 0:
            logical_offset, physical_offset, run_length, device = self.runs[run_idx]
            offset_in_run = offset - logical_offset
            remaining_in_run = run_length - offset_in_run

            read_length = min(length, remaining_in_run)
            device.fh.seek(physical_offset + offset_in_run)
            result.append(device.fh.read(read_length))

            length -= read_length
            offset += read_length
            run_idx += 1

        return b"".join(result)



from __future__ import annotations

from dissect.cstruct import cstruct

vmfs_def = """
/* === Common types === */
struct UUID {
    uint32          timeLo;
    uint32          timeHi;
    uint16          rand;
    char            macAddr[6];
};

struct SCSI_DiskId {
    uint8           type;
    uint8           len;
    uint16          lun;
    uint8           deviceType;
    uint8           scsiLevel;
    char            vendor[8];
    char            model[16];
    char            revision[4];
    char            id[44];
};

// These types normally have bitfields but we parse those in Python instead (address.py)
typedef uint32 FS3_Address;
typedef uint32 FS3_FileBlockAddr;
typedef uint32 FS3_FDAddr;

struct FS3_VolAddress {
    uint64 offset;
};

enum FS3_DescriptorType : uint32 {
    /* FS3DESC_... */
    INVALID,
    VOLUME,
    DIRECTORY,
    REGFILE,
    SYMLINK,
    SYSFILE,
    RDM,
    SENTINEL
};

enum FS3_AddrType {
    /* FS3_ADDR_... */
    INVALID             = 0x0,
    FILE_BLOCK          = 0x1,
    SMALL_FILE_BLOCK    = 0x1,                  /* File blocks, but on VMFS6 */
    SUB_BLOCK           = 0x2,
    POINTER_BLOCK       = 0x3,
    FILE_DESCRIPTOR     = 0x4,
    POINTER2_BLOCK      = 0x5,
    JOURNAL_BLOCK       = 0x6,
    LARGE_FILE_BLOCK    = 0x7,
    SENTINEL            = 0x8,
};

/* Added, don't have a better name */
enum FS3_ZeroLevelAddrType {
    INVALID             = 0x0,
    FILE_BLOCK          = 0x1,
    SUB_BLOCK           = 0x2,
    POINTER_BLOCK       = 0x3,
    POINTER2_BLOCK      = 0x5,
    POINTER_BLOCK_DOUBLE = 0x10D0,                  /* Don't have a better name */
    FILE_DESCRIPTOR_RESIDENT = 0x10D1,              /* Don't have a better name */
};

/* === System file addresses === */

#define rootDirDescAddr                 0x00000004
#define fbbDescAddr                     0x00400004
#define fdbcDescAddr                    0x00800004
#define pbcDescAddr                     0x00C00004
#define sbDescAddr                      0x01000004
#define vhDescAddr                      0x01400004
#define pb2DescAddr                     0x01800004
#define sdDirDescAddr                   0x01C00004
#define jbDescAddr                      0x02000004
#define unmapDescAddr                   0x02400004
#define dfDirDescAddr                   0x02800004

/* === Filesystem descriptor === */

#define FS3_FS_HEADER_OFFSET            0x200000
#define FS3_VMFS6_MAJOR_VERSION         24

#define VMFS_MAGIC_NUMBER               0x2fabf15e
#define VMFSL_MAGIC_NUMBER              0x2fabf15f
#define VMFS6_MAGIC_NUMBER              0x2fabf160  /* used? from symbols */
#define VMFS6L_MAGIC_NUMBER             0x2fabf161

flag FS3_Config : uint32 {
    UNKNOWN1            = 0x00000001,
    UNKNOWN2            = 0x00000002,
    DENSE_SBPC          = 0x00000004,               /* /config/VMFS3/intOpts/DenseSBPerCluster */
    MAINTENANCE         = 0x00000008,
    UNKNOWN10           = 0x00000010,               /* local VsanD? */
    UNKNOWN20           = 0x00000020,               /* default? */
    UNKNOWN40           = 0x00000040,               /* vmkfstools -C vmfs6 <dev> --createResourceOptimizedFS ? */
    UNKNOWN80           = 0x00000080,
    LOCAL               = 0x00000100,
    UNKNOWN200          = 0x00000200,               /* default? */
    SHARED              = 0x00000400,               /* vmkfstools -C vmfs6 <dev> --isVsanD */
    OSDATA              = 0x00000800,               /* vmkfstools -C vmfs6l <dev> --isSystem */
    VSAN                = 0x00001000,
    UNKNOWN2000         = 0x00002000,               /* default? */
    UNKNOWN4000         = 0x00004000,
};

struct FDS_VolInfo {
    char            id[32];
};

struct FS3_Checksum {
    uint64          value;
    uint64          checksumGen;
};

enum FS3_UnmapPriority : uint16 {
    NONE                = 0,
    LOW                 = 1,
    MEDIUM              = 2,
    HIGH                = 3,
};

enum FS3_UnmapMethod : uint8 {
    PRIORITY            = 0,
    FIXED               = 1,
};

struct FS3_Descriptor {
    uint32          magic;                          /* 0x00 */
    uint32          majorVersion;                   /* 0x04 */
    uint8           minorVersion;                   /* 0x08 */
    UUID            uuid;                           /* 0x09 */
    FS3_Config      config;                         /* 0x19 */
    char            fsLabel[128];                   /* 0x1d */
    uint32          diskBlockSize;                  /* 0x9d */
    uint64          fileBlockSize;                  /* 0xa1 */
    uint32          creationTime;                   /* 0xa9 */
    uint32          snapID;                         /* 0xad */
    FDS_VolInfo     volInfo;                        /* 0xb1 */
    uint32          fdcClusterGroupOffset;          /* 0xd1 */
    uint32          fdcClustersPerGroup;            /* 0xd5 */
    uint32          subBlockSize;                   /* 0xd9 */
    uint32          maxJournalSlotsPerTxn;          /* 0xdd */
    uint64          pb2VolAddr;                     /* 0xe1 */
    uint32          pb2FDAddr;                      /* 0xe9 */
    char            hostUuid[16];                   /* 0xed */
    uint64          gblGeneration;                  /* 0xfd */
    uint64          sddVolAddr;                     /* 0x105 */
    uint32          sddFDAddr;                      /* 0x10d */
    uint8           checksumType;                   /* 0x111 */
    FS3_UnmapPriority   unmapPriority;              /* 0x112 */
    char            pad1[4];                        /* 0x114 */
    uint64          checksumGen;                    /* 0x118 */
    FS3_Checksum    checksum;                       /* 0x120 */
    uint32          physDiskBlockSize;              /* 0x130 */
    uint32          mdAlignment;                    /* 0x134 */
    uint16          sfbToLfbShift;                  /* 0x138 */
    uint16          reserved16_1;                   /* 0x13a */
    uint16          reserved16_2;                   /* 0x13c */
    uint16          ptrBlockShift;                  /* 0x13e */
    uint16          sfbAddrBits;                    /* 0x140 */
    uint16          reserved16_3;                   /* 0x142 */
    uint32          tbzGranularity;                 /* 0x144 Also called unmapGranularity*/
    uint32          journalBlockSize;               /* 0x148 */
    uint32          leaseIntervalMs;                /* 0x14c */
    uint32          reclaimWindowMs;                /* 0x150 */
    uint64          localStampUS;                   /* 0x154 */
    char            localMountOwnerMacAddr[6];      /* 0x15c */
    FS3_UnmapMethod unmapMethod;                    /* 0x162 */
    uint32          unmapBandwidth;                 /* 0x163 */
    uint32          unmapBandwidthMin;              /* 0x167 */
    uint32          unmapBandwidthMax;              /* 0x16b */
};

/* === Resource metadata === */

enum FS3_ResourceTypeID {
    /* FS3_RESTYPE_... */
    INVALID             = 0,
    FILE_BLOCK          = 1,
    SMALL_FILE_BLOCK    = 1,
    SUB_BLOCK           = 2,
    PTR_BLOCK           = 3,
    FILE_DESC           = 4,
    PTR2_BLOCK          = 5,                        /* Don't have a better name */
    JOURNAL_BLOCK       = 6,
    LARGE_FILE_BLOCK    = 7,
    SENTINEL,
};

enum Res3PriorityType {
    /* RES3_PRIORITY_... */
    DEFAULT,
    JOURNAL             = 100000,
    MAX                 = 0x7fffffff
};

#define FS3_RFMD_SIGNATURE              0x72666D64  /* rfmd */
#define FS3_RCMD_SIGNATURE              0x72636D64  /* rcmd */

struct FS3_ResFileMetadata {
    uint32          resourcesPerCluster;            /* 0x00 */
    uint32          clustersPerGroup;               /* 0x04 */
    uint32          clusterGroupOffset;             /* 0x08 */
    uint32          resourceSize;                   /* 0x0c */
    uint32          clusterGroupSize;               /* 0x10 */
    uint32          numResourcesLo;                 /* 0x14 */
    uint32          numClusterGroups;               /* 0x18 */
    uint32          numResourcesHi;                 /* 0x1c */
    uint32          signature;                      /* 0x20 */
    uint32          version;                        /* 0x24 */
    uint32          flags;                          /* 0x28 0x02 = has parent?, 0x20 = HasUnmapBitmap */
    uint16          affinityPerCluster;             /* 0x2c */
    uint16          affinityPerResource;            /* 0x2e */
    uint32          bitsPerResource;                /* 0x30 */
    uint32          childMetaOffset;                /* 0x34 */
    uint32          parentResourcesPerCluster;      /* 0x38 */
    uint32          parentClustersPerGroup;         /* 0x3c */
    uint32          parentClusterGroupSize;         /* 0x40 */
    char            _unknown[16];                   /* 0x44 */
    uint32          convertedClusters;              /* 0x54 */
};

struct FS3_ResourceClusterMD {
    uint32          clusterNum;
    uint32          totalResources;
    uint32          freeResources;
    uint32          nextFreeIdx;
    uint8           bitmap[64];
    uint8           typeData[16];
    Res3PriorityType    priority;
};

/* TODO: This is not fully researched yet */
struct Res3_ClusterMetaVMFS6 {
    char            _pad[16];                       /* 0x00 */
    uint32          signature;                      /* 0x10 */
    uint32          priority;                       /* 0x14 */
    uint64          clusterNum;                     /* 0x18 */
    uint16          totalResources;                 /* 0x20 */
    uint16          freeResources;                  /* 0x22 */
    uint16          unk1;                           /* 0x24 */
    uint16          affinityCount;                  /* 0x26 */
    char            _unk[2328];                     /* 0x28 */
    /* List of tuple of addresses start here, referenced as overflow keys */
};

/* === Disk lock and heartbeat === */
/* TODO: This is not fully researched yet */

struct FS3_HBGen {
    uint64          gen;
};

struct FS3_LockHolder {
    FS3_VolAddress  hbAddr;
    FS3_HBGen       hbGen;
    UUID            uid;
};

struct FS3_DiskLock {
    uint32          type;                           /* 0x00 */
    FS3_VolAddress  addr;                           /* 0x04 */
    FS3_VolAddress  hbAddr;                         /* 0x0c */
    FS3_HBGen       hbGen;                          /* 0x14 */
    uint64          token;                          /* 0x1c */
    uint32          mode;                           /* 0x24 */
    UUID            owner;                          /* 0x28 */
    uint64          mtime;                          /* 0x38 */
    uint32          numHolders;                     /* 0x40 */
    FS3_LockHolder  holders[8];                     /* 0x44 */
    char            _unk[32];                       /* 0x144 */
    uint32          gblNumHolders;                  /* 0x164 */
    uint64          gblGen;                         /* 0x168 */
    uint32          gblBrk;                         /* 0x170 */
};

struct FS3_Heartbeat {
    uint32          state;                          /* 0x00 */
    FS3_VolAddress  addr;                           /* 0x04 */
    FS3_HBGen       hbGen;                          /* 0x0c */
    uint64          stamp;                          /* 0x14 */
    UUID            owner;                          /* 0x1c */
    FS3_Address     journalAddr;                    /* 0x2c */
    uint32          driveX;                         /* 0x30 drv %u.%u */
    uint8           driveY;                         /* 0x34 */
    uint8           lockImpl;                       /* 0x35 */
    char            ip[46];                         /* 0x36 */
    uint64          journalVolAddr;                 /* 0x64 */
    UUID            replayHost;                     /* 0x6c */
    char            _unk[24];                       /* 0x7c */
    uint64          replayHostHB;                   /* 0x94 */
    FS3_HBGen       replayHostHBgen;                /* 0x9c */
};

/* === FD/inode info === */

#define FS3_FDMD_SIGNATURE              0x66646D64  /* fdmd */

struct FS3_FileMetadata {
    FS3_FDAddr      descAddr;                       /* 0x00 */
    uint32          generation;                     /* 0x04 */
    uint32          linkCount;                      /* 0x08 */
    FS3_DescriptorType  type;                       /* 0x0c */
    uint32          flags;                          /* 0x10 1 = Large file allocation? */
    uint64          fileLength;                     /* 0x14 */
    uint64          blockSize;                      /* 0x1c */
    uint64          numBlocks;                      /* 0x24 */
    uint32          mtime;                          /* 0x2c */
    uint32          ctime;                          /* 0x30 */
    uint32          atime;                          /* 0x34 */
    uint32          uid;                            /* 0x38 */
    uint32          gid;                            /* 0x3c */
    uint32          mode;                           /* 0x40 */
    uint32          zeroLevelAddrType;              /* 0x44 */
    uint32          numTBZBlocksLo;                 /* 0x48 */
    uint32          numCOWBlocksLo;                 /* 0x4c */
    uint32          newSinceEpochLo;                /* 0x50 */
    uint32          numTBZBlocksHi;                 /* 0x54 */
    uint32          numCOWBlocksHi;                 /* 0x58 */
    uint32          numPointerBlocks;               /* 0x5c */
    uint32          newSinceEpochHi;                /* 0x60 */
    uint32          signature;                      /* 0x64 */
    uint32          affinityFD;                     /* 0x68 */
    uint32          tbzGranularityShift;            /* 0x6c */
    uint32          parentFD;                       /* 0x70 */
    uint32          lastSFBClusterNum;              /* 0x74 */
    uint32          _unk4;                          /* 0x78 */
    uint32          _unk5;                          /* 0x7c */
    uint32          affinityBitmap;                 /* 0x80 */
    uint8           numPreAllocBlocks;              /* 0x84 */
    uint8           _unk7;                          /* 0x85 */
    uint8           _unk8;                          /* 0x86 */
    uint8           _unk9;                          /* 0x87 */
    uint8           _unk10;                         /* 0x88 */
    uint8           blockOffsetShift;               /* 0x89 */
    uint8           numTracked;                     /* 0x8a */
    uint8           _unk12;                         /* 0x8b */
    uint32          numLFB;                         /* 0x8c */
    char            _unk13[216];                    /* 0x90 */
    uint32          lastFreeSFBC;                   /* 0x168 */
};

struct FS3_RawDiskMap {
    SCSI_DiskId     diskId;
};

/* === Directory entries === */

#define FS3_MAX_FILE_NAME_LENGTH        128
#define FS6_MAX_FILE_NAME_LENGTH        256

struct FS3_DirEntry {
    FS3_DescriptorType  type;                       /* 0x00 */
    FS3_FDAddr      descAddr;                       /* 0x04 */
    uint32          generation;                     /* 0x08 */
    char            name[FS3_MAX_FILE_NAME_LENGTH]; /* 0x0c */
};                                                  /* 0x8c */

#define FS6_DIR_HEADER_VERSION          0xF50001
#define FS6_DIR_HEADER_DEBUG_VERSION    0xFDC001
#define FS6_DIR_HEADER_BLOCK_SIZE       0x10000

#define FS6_DIR_HASH_MAX_ENTRIES        16001
#define FS6_DIR_HASH_ROOT_RESERVED      28
#define FS6_DIR_HASH_MAX_ROOT_ENTRIES   (FS6_DIR_HASH_MAX_ENTRIES - FS6_DIR_HASH_ROOT_RESERVED)

#define FS6_DIR_LINKS_PER_GROUP         12

enum FS6_DirBlockType {
    DIRENT              = 0x1,
    LINK                = 0x2,
    ALLOCATION_MAP      = 0x3,
};

struct FS6_DirEntry {
    FS3_DescriptorType  type;                       /* 0x00 */
    uint32          address;                        /* 0x04 */
    uint32          generation;                     /* 0x08 */
    uint16          hashIndex;                      /* 0x0c */
    uint16          linkHash;                       /* 0x0e */
    uint64          offset;                         /* 0x10 */
    char            name[256];                      /* 0x18 */
    uint64          _unk1;                          /* 0x118 */
};                                                  /* 0x120 */

struct FS6_DirHeader {
    uint32          version;                        /* 0x00 */
    uint32          numEntries;                     /* 0x04 */
    uint32          numAllocated;                   /* 0x08 */
    uint32          numAllocationMapBlocks;         /* 0x0c */
    uint32          allocationMapBlocks[128];       /* 0x10 */
    char            _gap0[408];                     /* 0x210 */
    uint64          checksum;                       /* 0x3A8 */
    uint64          checksumGen;                    /* 0x3B0 */
    FS6_DirEntry    selfEntry;                      /* 0x3B8 */
    FS6_DirEntry    parentEntry;                    /* 0x4D8 */
};                                                  /* 0x5F8 */

struct FS6_DirBlockHeader {
    uint16          version;                        /* 0x00 */
    uint16          type;                           /* 0x02 */
    uint16          totalSlots;                     /* 0x04 */
    uint16          freeSlots;                      /* 0x06 */
    char            bitmap[56];                     /* 0x08 */
};                                                  /* 0x40 */

struct FS6_DirLink {
    uint32          location;                       /* 0x00 */
    uint16          hash;                           /* 0x04 */
};                                                  /* 0x06 */

struct FS6_DirLinkGroup {
    uint16          hashIndex;                      /* 0x00 The hash index this is a link for */
    uint8           totalLinks;                     /* 0x02 */
    uint8           freeLinks;                      /* 0x03 */
    uint8           nextFreeIdx;                    /* 0x04 */
    uint8           pad1;                           /* 0x05 */
    FS6_DirLink     links[FS6_DIR_LINKS_PER_GROUP]; /* 0x06 */
    uint32          nextGroup;                      /* 0x4e */
    uint16          pad2;                           /* 0x52 */
};                                                  /* 0x54 */
"""

c_vmfs = cstruct().load(vmfs_def)

FS3_Config = c_vmfs.FS3_Config
FS3_AddrType = c_vmfs.FS3_AddrType
FS3_ResourceTypeID = c_vmfs.FS3_ResourceTypeID
FS3_ZeroLevelAddrType = c_vmfs.FS3_ZeroLevelAddrType
FS3_DescriptorType = c_vmfs.FS3_DescriptorType
FS6_DirBlockType = c_vmfs.FS6_DirBlockType



from __future__ import annotations

from dissect.cstruct import cstruct

lvm_def = """
typedef uint8 Bool;

#define DISK_BLOCK_SIZE_512B            512
#define FS_PLIST_DEF_MAX_PARTITIONS     32

#define LVM_MAGIC_NUMBER                0xC001D00D
#define LVM_INITIAL_MAJOR_VERSION       3
#define LVM_MAJOR_VERSION_ESX60         6
#define LVM_MAJOR_VERSION_ESX50         5

#define LVM_IS_LVM6(majorVersion)       ((majorVersion == LVM_MAJOR_VERSION_ESX60) ? TRUE : FALSE)

#define LVM_DEV_HEADER_OFFSET           0x00100000
#define LVM_MD_ALIGNMENT_DEFAULT        (1 << 12)

#define LVM_MAX_VOLUME_LABEL_LENGTH     64

#define LVM_SIZEOF_LVM_DEVMETA_LVM5     DISK_BLOCK_SIZE_512B
#define LVM_SIZEOF_LVM_DEVMETA_LVM6(mdAlignment)    (mdAlignment)
#define LVM_SIZEOF_LVM_DEVMETA(majorVersion,mdAlignment) (LVM_IS_LVM6(majorVersion) ? LVM_SIZEOF_LVM_DEVMETA_LVM6(mdAlignment) : LVM_SIZEOF_LVM_DEVMETA_LVM5)

#define LVM_SIZEOF_EXTVOLMETA           256
#define LVM_SIZEOF_VTENTRY              512
#define LVM_SIZEOF_PTENTRY              128
#define LVM_SIZEOF_SDTENTRY             256

#define LVM_MAX_VOLUMES_PER_DEV_LVM5    512
#define LVM_MAX_VOLUMES_PER_DEV_LVM6    1
#define LVM_MAX_VOLUMES_PER_DEV(majorVersion) (LVM_IS_LVM6(majorVersion) ? LVM_MAX_VOLUMES_PER_DEV_LVM6 : LVM_MAX_VOLUMES_PER_DEV_LVM5)

#define LVM_UNUSED_MD_SECTORS_LVM5      (1024 - (LVM_MAX_VOLUMES_PER_DEV_LVM5))
#define LVM_UNUSED_MD_SECTORS_LVM6      (1024 - (LVM_MAX_VOLUMES_PER_DEV_LVM6))

#define LVM_UNUSED_MD_SIZE_LVM5         LVM_UNUSED_MD_SECTORS_LVM5 * DISK_BLOCK_SIZE_512B
#define LVM_UNUSED_MD_SIZE_LVM6         LVM_UNUSED_MD_SECTORS_LVM6 * DISK_BLOCK_SIZE_512B

#define LVM_RESERVED_SIZE_LVM5          (LVM_UNUSED_MD_SIZE_LVM5 - LVM_SIZEOF_SDTENTRY * FS_PLIST_DEF_MAX_PARTITIONS)
#define LVM_RESERVED_SIZE_LVM6          (LVM_UNUSED_MD_SIZE_LVM6 - LVM_SIZEOF_SDTENTRY * FS_PLIST_DEF_MAX_PARTITIONS)
#define LVM_RESERVED_SIZE(majorVersion) (LVM_IS_LVM6(majorVersion) ? LVM_RESERVED_SIZE_LVM6 : LVM_RESERVED_SIZE_LVM5)

#define LVM_PES_PER_BITMAP              8192
#define LVM_PE_BITMAP_SIZE_LVM5         (LVM_PES_PER_BITMAP / 8)
#define LVM_PE_BITMAP_SIZE_LVM6(mdAlignment)    (MAX(mdAlignment, LVM_PE_BITMAP_SIZE_LVM5))
#define LVM_PE_BITMAP_SIZE(majorVersion,mdAlignment)    (LVM_IS_LVM6(majorVersion) ? LVM_PE_BITMAP_SIZE_LVM6(mdAlignment) : LVM_PE_BITMAP_SIZE_LVM5)

enum LVM_VolState {
    LVM_INVALID,
    LVM_NORMAL,
    LVM_RESIG_PENDING,
    LVM_RESIG_DONE,
    LVM_SENTINEL
};

struct UUID {
    uint32          timeLo;
    uint32          timeHi;
    uint16          rand;
    char            macAddr[6];
};

struct SCSI_DiskId {
    uint8           type;
    uint8           len;
    uint16          lun;
    uint8           deviceType;
    uint8           scsiLevel;
    char            vendor[8];
    char            model[16];
    char            revision[4];
    char            id[44];
};

struct LVM_DevMetadata {
    uint32          magic;                              /* 0x00 0xC001D00D */
    uint32          majorVersion;                       /* 0x04 */
    uint32          minorVersion;                       /* 0x08 */
    SCSI_DiskId     diskID;                             /* 0x0C */
    uint32          diskBlockSize;                      /* 0x5a */
    uint64          totalBytes;                         /* 0x5e */
    uint32          numVolumes;                         /* 0x66 */
    uint32          numPEs;                             /* 0x6a */
    uint32          lastPEIndex;                        /* 0x6e */
    uint64          generation;                         /* 0x72 */
    uint64          dataOffset;                         /* 0x7A */
    UUID            devID;                              /* 0x82 */
    uint64          initTimeUS;                         /* 0x92 */
    uint64          modTimeUS;                          /* 0x9A */
    UUID            lockedBy;                           /* 0xA2 */
    uint64          lockedWhenUS;                       /* 0xB2 */
    uint32          _unknown0;                          /* 0xBA */
    uint32          numPEMaps;                          /* 0xBE */
    uint64          extDevMetadataOffset;               /* 0xC2 */
    uint32          mdAlignment;                        /* 0xCA Not sure if this is a good name, but it's used similarly */
    uint32          numPEs6;                            /* 0xCE */
    uint32          flags;                              /* 0xD2 1 == VMFSL */
};                                                      /* 0xD6 sizeof(LVM_DevMetadata) */

struct LVM_ExtDevMetadata {
    uint32          magic;                              /* 0x00 0xC001D00D */
    uint32          numPEMaps;                          /* 0x04 */
    uint64          dataOffset;                         /* 0x08 */
    uint64          nextOffset;                         /* 0x10 Offset to the next LVM_ExtDevMetadata */
};                                                      /* 0x18 sizeof(LVM_ExtDevMetadata) */

struct LVM_VolID {
    UUID            uuid;                               /* 0x00 UUID of the volume */
    uint32          snapID;                             /* 0x10 Snapshot ID of the volume */
};                                                      /* 0x14 sizeof(LVM_VolID) */

struct LVM_VolMetadata {
    uint64          logicalSize;                        /* 0x00 Logical size of the volume in bytes */
    uint64          generation;                         /* 0x08 Generation number of the volume */
    LVM_VolState    state;                              /* 0x10 State of the volume */
    char            name[LVM_MAX_VOLUME_LABEL_LENGTH];  /* 0x14 Name of the volume */
    LVM_VolID       lvID;                               /* 0x54 Logical volume ID */
    uint64          creationTimeUS;                     /* 0x64 Creation time of the volume in microseconds */
};                                                      /* 0x70 sizeof(LVM_VolMetadata) */

struct LVM_ExtVolMetadata {
    uint32          numDevs;                            /* 0x00 Number of devices in the volume */
    char            pad[124];                           /* 0x04 */
    uint32          consumedPEs;                        /* 0x80 Consumed physical extents */
};                                                      /* 0x84 sizeof(LVM_ExtVolMetadata) */

struct LVM_VolDescriptor {
    LVM_VolMetadata volMeta;                            /* 0x00 */
    uint32          volumeID;                           /* 0x70 Volume index */
    uint32          numPEs;                             /* 0x74 Number of physical extents */
    uint64          firstPE;                            /* 0x78 First physical extent */
    uint64          lastPE;                             /* 0x80 Last physical extent */
    uint64          modTimeUS;                          /* 0x88 Last modification time in microseconds */
    LVM_ExtVolMetadata  extVolMeta;                     /* 0x90 Extended volume metadata */
    char            pad[236];                           /* 0x114 */
};                                                      /* 0x200 sizeof(LVM_VolDescriptor) */

struct LVM_VolTableEntry {
    LVM_VolDescriptor   volDesc;                        /* 0x00 */
};                                                      /* 0x200 sizeof(LVM_VolTableEntry) */

struct LVM_PEDescriptor {
    uint32          peID;                               /* 0x00 Physical extent ID */
    uint32          volumeID;                           /* 0x04 Volume ID this PE belongs to */
    uint64          pOffset;                            /* 0x08 Physical offset of this PE in the device */
    uint64          lOffset;                            /* 0x10 Logical offset of this PE in the volume */
    uint64          length;                             /* 0x18 Length of this PE in bytes */
    uint32          version;                            /* 0x20 Version of this PE */
};                                                      /* 0x24 sizeof(LVM_PEDescriptor) */

struct LVM_PETableEntry {
    Bool            used;                               /* 0x00 Whether this PE is used */
    LVM_PEDescriptor    peDesc;                         /* 0x01 Physical extent descriptor */
};                                                      /* 0x25 sizeof(LVM_PETableEntry) */

struct LVM_SDTableEntry {
    char            deviceName[256];                    /* 0x00 Name of the device */
};                                                      /* 0x100 sizeof(LVM_SDTableEntry) */
"""  # noqa: E501

c_lvm = cstruct().load(lvm_def)



from __future__ import annotations

from typing import ClassVar

from dissect.vmfs.c_vmfs import FS3_AddrType


def address_type(address: int) -> int:
    """Return the address type.

    Address type is encoded in the lower 3 bits.
    """
    return address & 0b111


def address_fmt(address: int) -> str:
    return repr(Address(address))


class Address:
    """Base class for VMFS addresses.

    This class primarily exists to provide an easy way to inspect addresses when debugging or testing interactively.
    For every other case where we actually parse addresses, we use the static methods of the subclasses.
    This is primarily for performance reasons, so we avoid the overhead of creating unnecessary class instances
    when we just want one or two integers.

    VMFS makes extensive use of opaque 32-bit and 64-bit integers to represent addresses.
    In VMFS5, exclusively 32-bit integers are used, while VMFS6 uses 64-bit integers for most address types.
    The lower 3 bits of an address determine the address type, while the remaining bits are address type specific.
    Some address types have additional bits for flags, such as copy-on-write (COW) or to-be-zeroed (TBZ).
    Most address types have a cluster and resource part, which are used to index into resource files.
    The file block and large file block addresses only have a block part, which points to somewhere on the volume.

    .. rubric :: Encoding
    .. code-block:: c

        struct FS3_Address {
            uint32 addrType : 3;
            uint32 addrSpecific : 29;
        };

        struct FS3_Address64 {
            uint64 addrType : 3;
            uint64 addrSpecific : 61;
        };

    References:
        - ``Addr3_AddrToStr``
        - ``Addr3_AddrToStr64``
        - ``Res3Parse*Addr*``
    """

    __type__ = FS3_AddrType.INVALID
    __known_types__: ClassVar[dict[int, type[Address]]] = {}

    def __init_subclass__(cls):
        cls.__known_types__[cls.__type__.value] = cls

    def __new__(
        cls, value: int = 0
    ) -> (
        Address
        | FileBlockAddr
        | SmallFileBlockAddr
        | SubBlockAddr
        | SubBlockAddr64
        | PointerBlockAddr
        | PointerBlockAddr64
        | FileDescriptorAddr
        | PointerBlock2Addr
        | PointerBlock2Addr64
        | JournalBlockAddr
        | LargeFileBlockAddr
    ):
        """Create an instance of the appropriate address type based on the value."""
        if cls is Address:
            if (addr_type := address_type(value)) not in cls.__known_types__:
                return super().__new__(cls)
            return cls.__known_types__[addr_type](value)
        return super().__new__(cls)

    def __init__(self, value: int = 0):
        self.value = value

    def __int__(self) -> int:
        return self.value

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Address):
            return self.value == other.value
        if isinstance(other, int):
            return self.value == other
        return False

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self.value)

    def __repr__(self) -> str:
        if self.value == 0:
            return "<Null address>"
        return f"<Invalid address {self.value:#x}>"

    @property
    def type(self) -> int:
        """Return the address type."""
        return address_type(self.value)

    @staticmethod
    def parse(address: int) -> tuple[int, int] | int:
        """Parse the address and return the cluster and resource.

        This method should be overridden by subclasses.
        """
        raise NotImplementedError("Subclasses must implement parse method.")

    @classmethod
    def make(cls, cluster: int, resource: int) -> int:
        """Create an address from a cluster and resource.

        This method should be overridden by subclasses.
        """
        raise NotImplementedError("Subclasses must implement make method.")

    @classmethod
    def make_type(cls, type: FS3_AddrType | int, cluster: int, resource: int) -> int:
        """Create an address from a type, cluster and resource."""
        return cls.__known_types__[int(type)].make(cluster, resource)


class FileBlockAddr(Address):
    """VMFS5 FB address.

    FB (file block) in VMFS5 are used to describe data blocks on the volume.

    A block in a lazily zeroed file, can have its TBZ (to-be-zeroed) flag set to 1, indicating that the block should be
    treated as zeroed when read. This is treated for the whole block.

    .. rubric :: Encoding
    .. code-block:: c

        struct FS3_FileBlockAddr {
            uint32 addrType     : 3;
            uint32 unused       : 1;
            uint32 copyOnWrite  : 1;
            uint32 toBeZeroed   : 1;
            uint32 blockNum     : 26;
        };

        0b00000000 00000000 00000000 00000111  (type)
        0b00000000 00000000 00000000 00001000  (unused/unknown)
        0b00000000 00000000 00000000 00010000  (cow)
        0b00000000 00000000 00000000 00100000  (tbz)
        0b11111111 11111111 11111111 11000000  (block)
    """

    __type__ = FS3_AddrType.FILE_BLOCK

    def __repr__(self) -> str:
        return f"<FB tbz={self.tbz} cow={self.cow} block={self.block}>"

    @property
    def cow(self) -> int:
        """Parse the FB address and return the copy-on-write value."""
        return (self.value >> 4) & 0x1

    @property
    def tbz(self) -> int:
        """Parse the FB address and return the to-be-zeroed value."""
        return (self.value >> 5) & 0x1

    @property
    def block(self) -> int:
        return self.parse(self.value)

    @staticmethod
    def parse(address: int) -> int:
        return (address >> 6) & 0x3FFFFFF

    @classmethod
    def make(
        cls,
        block: int,
        cow: int = 0,
        tbz: int = 0,
    ) -> int:
        """Create a VMFS5 FB address from a block."""
        return cls.__type__ | ((cow & 0x1) << 4) | ((tbz & 0x1) << 5) | ((block & 0x3FFFFFF) << 6)


class SmallFileBlockAddr(Address):
    """VMFS6 SFB address.

    SFB (small file block) in VMFS6 is what FB was in VMFS5. They are used to describe data blocks on the volume.
    Their encoding now consists of a cluster and resource part, which can be used to check allocation status
    in the resource file. See :class:`~dissect.vmfs.resource.SmallFileBlockResourceVMFS6` for more details.

    In this variant of the address, the TBZ (to-be-zeroed) flag is a bitmap that can be used to indicate
    different parts of the block that should be zeroed out. Though when reading the block, the whole block seems to
    be treated as zeroed when any bit is set to 1.

    .. rubric :: Encoding
    .. code-block:: c

        struct FS3_SmallFileBlockAddr64 {
            uint64 addrType         : 3;
            uint64 unused1          : 2;
            uint64 copyOnWrite      : 1;
            uint64 unused2          : 1;
            uint64 toBeZeroed       : 8;
            uint64 sfbClusterNum    : 31;
            uint64 unknown          : 5;
            uint64 sfbNum           : 13;
        };

        0b00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000111  (type)
        0b00000000 00000000 00000000 00000000 00000000 00000000 00000000 00011000  (unused/unknown)
        0b00000000 00000000 00000000 00000000 00000000 00000000 00000000 00100000  (cow)
        0b00000000 00000000 00000000 00000000 00000000 00000000 00000000 01000000  (unused/unknown)
        0b00000000 00000000 00000000 00000000 00000000 00000000 01111111 10000000  (tbz bitmap)
        0b00000000 00000000 00111111 11111111 11111111 11111111 10000000 00000000  (cluster)
        0b00000000 00000111 11000000 00000000 00000000 00000000 00000000 00000000  (unknown)
        0b11111111 11111000 00000000 00000000 00000000 00000000 00000000 00000000  (resource)
    """

    __type__ = FS3_AddrType.SMALL_FILE_BLOCK

    def __repr__(self) -> str:
        return f"<SFB tbz={self.tbz:#x} cow={self.cow} c{self.cluster} r{self.resource}>"

    @property
    def cow(self) -> int:
        """Parse the SFB address and return the copy-on-write value."""
        return (self.value >> 5) & 0x1

    @property
    def tbz(self) -> int:
        """Parse the SFB address and return the to-be-zeroed value."""
        return (self.value >> 7) & 0xFF

    @property
    def cluster(self) -> int:
        """Return the cluster part of the address."""
        cluster, _ = self.parse(self.value)
        return cluster

    @property
    def resource(self) -> int:
        """Return the resource part of the address."""
        _, resource = self.parse(self.value)
        return resource

    @staticmethod
    def parse(address: int) -> tuple[int, int]:
        """Parse the SFB address and return the cluster and resource."""
        cluster = (address >> 15) & 0x7FFFFFFF
        resource = (address >> 51) & 0x1FFF
        return cluster, resource

    @classmethod
    def make(cls, cluster: int, resource: int, cow: int = 0, tbz: int = 0) -> int:
        """Create a VMFS6 SFB address from a cluster and resource."""
        return (
            cls.__type__
            | ((cow & 0x1) << 5)
            | ((tbz & 0xFF) << 7)
            | ((cluster & 0x7FFFFFFF) << 15)
            | ((resource & 0x1FFF) << 51)
        )


class SubBlockAddr(Address):
    """VMFS5 SB address.

    SB (sub-block) describe small blocks of data which would be wasteful to store in a normal file block, and instead
    are stored in the ``.sbc.sf`` resource file.

    If the ``/config/VMFS3/intOpts/DenseSBPerCluster`` option is set, the resource part of the address
    is extended by 2 bits to allow more sub-block resources per cluster.

    .. rubric :: Encoding
    .. code-block:: c

        struct FS3_SubBlockAddr {
            uint32 addrType     : 3;
            uint32 sbNumHi      : 2;
            uint32 copyOnWrite  : 1;
            uint32 sbClusterNum : 22;
            uint32 sbNum        : 4;
        };

        0b00000000 00000000 00000000 00000111  (type)
        0b00000000 00000000 00000000 00011000  (resourceHi if FS3_Config.DENSE_SBPC is set)
        0b00000000 00000000 00000000 00100000  (cow)
        0b00001111 11111111 11111111 11000000  (cluster)
        0b11110000 00000000 00000000 00000000  (resource)
    """

    __type__ = FS3_AddrType.SUB_BLOCK

    def __repr__(self) -> str:
        return f"<SB cow={self.cow} c{self.cluster} r{self.resource}>"

    @property
    def cow(self) -> int:
        """Parse the SB address and return the copy-on-write value."""
        return (self.value >> 5) & 0x1

    @property
    def cluster(self) -> int:
        """Return the cluster part of the address."""
        cluster, _ = self.parse(self.value)
        return cluster

    @property
    def resource(self) -> int:
        """Return the resource part of the address."""
        _, resource = self.parse(self.value)
        return resource

    @staticmethod
    def parse(address: int, dense: bool = False) -> tuple[int, int]:
        """Parse the SB address and return the cluster and resource."""
        cluster = (address >> 6) & 0x3FFFFF
        resource = (address >> 28) & 0xF
        if dense:
            # In dense mode, an increased the number of sub-block resources are stored per cluster
            # /config/VMFS3/intOpts/DenseSBPerCluster
            resource |= (address & 0b11000) << 1
        return cluster, resource

    @classmethod
    def make(cls, cluster: int, resource: int, cow: int = 0) -> int:
        """Create a VMFS5 SB address from a cluster and resource."""
        return cls.__type__ | ((cow & 0x1) << 5) | ((cluster & 0x3FFFFF) << 6) | ((resource & 0xF) << 28)


class SubBlockAddr64(Address):
    """VMFS6 SB address.

    SB (sub-block) describe small blocks of data which would be wasteful to store in a normal file block, and instead
    are stored in the ``.sbc.sf`` resource file.

    .. rubric :: Encoding
    .. code-block:: c

        struct FS3_SubBlockAddr64 {
            uint64 addrType     : 3;
            uint64 unused       : 2;
            uint64 copyOnWrite  : 1;
            uint64 sbClusterNum : 36;
            uint64 unknown      : 14;
            uint64 sbNum        : 8;
        };

        0b00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000111  (type)
        0b00000000 00000000 00000000 00000000 00000000 00000000 00000000 00011000  (unused/unknown)
        0b00000000 00000000 00000000 00000000 00000000 00000000 00000000 00100000  (cow)
        0b00000000 00000000 00000000 11111111 11111111 11111111 11111111 11000000  (cluster)
        0b00000000 00000000 00000011 00000000 00000000 00000000 00000000 00000000  (cluster according to Addr3_AddrToStr64, but not Res3ParseSubBlockAddr64VMFS6)
        0b00000000 11111111 11111100 00000000 00000000 00000000 00000000 00000000  (unknown)
        0b11111111 00000000 00000000 00000000 00000000 00000000 00000000 00000000  (resource)
    """  # noqa: E501

    __type__ = FS3_AddrType.SUB_BLOCK

    def __repr__(self) -> str:
        return f"<SB cow={self.cow} c{self.cluster} r{self.resource}>"

    @property
    def cow(self) -> int:
        """Parse the SB address and return the copy-on-write value."""
        return (self.value >> 5) & 0x1

    @property
    def cluster(self) -> int:
        """Return the cluster part of the address."""
        cluster, _ = self.parse(self.value)
        return cluster

    @property
    def resource(self) -> int:
        """Return the resource part of the address."""
        _, resource = self.parse(self.value)
        return resource

    @staticmethod
    def parse(address: int) -> tuple[int, int]:
        """Parse the SB address and return the cluster and resource."""
        cluster = (address >> 6) & 0xFFFFFFFFF
        resource = (address >> 56) & 0xFF
        return cluster, resource

    @classmethod
    def make(cls, cluster: int, resource: int, cow: int = 0) -> int:
        """Create a VMFS6 SB address from a cluster and resource."""
        return cls.__type__ | ((cow & 0x1) << 5) | ((cluster & 0xFFFFFFFFF) << 6) | ((resource & 0xFF) << 56)


class PointerBlockAddr(Address):
    """VMFS5 PB/PB2 address.

    PB (pointer block) are used for indirect addressing of file blocks. They can point to other PBs or file blocks.

    .. rubric :: Encoding
    .. code-block:: c

        struct FS3_PtrBlockAddr {
            uint32 addrType     : 3;
            uint32 unused       : 2;
            uint32 copyOnWrite  : 1;
            uint32 pbClusterNum : 22;
            uint32 ptrBlockNum  : 4;
        };

        0b00000000 00000000 00000000 00000111  (type)
        0b00000000 00000000 00000000 00011000  (unused)
        0b00000000 00000000 00000000 00100000  (cow)
        0b00001111 11111111 11111111 11000000  (cluster)
        0b11110000 00000000 00000000 00000000  (resource)
    """

    __type__ = FS3_AddrType.POINTER_BLOCK

    def __repr__(self) -> str:
        return f"<PB cow={self.cow} c{self.cluster} r{self.resource}>"

    @property
    def cow(self) -> int:
        """Parse the PB address and return the copy-on-write value."""
        return (self.value >> 5) & 0x1

    @property
    def cluster(self) -> int:
        """Return the cluster part of the address."""
        cluster, _ = self.parse(self.value)
        return cluster

    @property
    def resource(self) -> int:
        """Return the resource part of the address."""
        _, resource = self.parse(self.value)
        return resource

    @staticmethod
    def parse(address: int) -> tuple[int, int]:
        """Parse the PB address and return the cluster and resource."""
        cluster = (address >> 6) & 0x3FFFFF
        resource = (address >> 28) & 0xF
        return cluster, resource

    @classmethod
    def make(cls, cluster: int, resource: int, cow: int = 0) -> int:
        """Create a VMFS5 PB address from a cluster and resource."""
        return cls.__type__ | ((cow & 0x1) << 5) | ((cluster & 0x3FFFFF) << 6) | ((resource & 0xF) << 28)


class PointerBlockAddr64(Address):
    """VMFS6 PB/PB2 address.

    PB (pointer block) are used for indirect addressing of file blocks. They can point to other PBs or file blocks.

    .. rubric :: Encoding
    .. code-block:: c

        struct FS3_PtrBlockAddr64 {
            uint64 addrType     : 3;
            uint64 unused       : 2;
            uint64 copyOnWrite  : 1;
            uint64 pbClusterNum : 36;
            uint64 unknown      : 14;
            uint64 ptrBlockNum  : 8;
        };

        0b00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000111  (type)
        0b00000000 00000000 00000000 00000000 00000000 00000000 00000000 00011000  (unused/unknown)
        0b00000000 00000000 00000000 00000000 00000000 00000000 00000000 00100000  (cow)
        0b00000000 00000000 00000011 11111111 11111111 11111111 11111111 11000000  (cluster)
        0b00000000 11111111 11111100 00000000 00000000 00000000 00000000 00000000  (unknown)
        0b11111111 00000000 00000000 00000000 00000000 00000000 00000000 00000000  (resource)
    """

    __type__ = FS3_AddrType.POINTER_BLOCK

    def __repr__(self) -> str:
        return f"<PB cow={self.cow} c{self.cluster} r{self.resource}>"

    @property
    def cow(self) -> int:
        """Parse the PB address and return the copy-on-write value."""
        return (self.value >> 5) & 0x1

    @property
    def cluster(self) -> int:
        """Return the cluster part of the address."""
        cluster, _ = self.parse(self.value)
        return cluster

    @property
    def resource(self) -> int:
        """Return the resource part of the address."""
        _, resource = self.parse(self.value)
        return resource

    @staticmethod
    def parse(address: int) -> tuple[int, int]:
        """Parse the PB address and return the cluster and resource."""
        cluster = (address >> 6) & 0xFFFFFFFFF
        resource = (address >> 56) & 0xFF
        return cluster, resource

    @classmethod
    def make(cls, cluster: int, resource: int, cow: int = 0) -> int:
        """Create a VMFS6 PB address from a cluster and resource."""
        return cls.__type__ | ((cow & 0x1) << 5) | ((cluster & 0xFFFFFFFFF) << 6) | ((resource & 0xFF) << 56)


class FileDescriptorAddr(Address):
    """FD address. Used in both VMFS5 and VMFS6.

    FD (file descriptor) addresses are used to reference file descriptors in the ``.fdc.sf`` resource file.
    They are similar to inode numbers in traditional filesystems.

    .. rubric :: Encoding
    .. code-block:: c

        struct FS3_FDAddr {
            uint32 addrType     : 3;
            uint32 unused       : 3;
            uint32 fdClusterNum : 16;
            uint32 fdNum        : 10;
        };

        0b00000000 00000000 00000000 00000111  (type)
        0b00000000 00000000 00000000 00111000  (unused)
        0b00000000 00111111 11111111 11000000  (cluster)
        0b11111111 11000000 00000000 00000000  (resource)
    """

    __type__ = FS3_AddrType.FILE_DESCRIPTOR

    def __repr__(self) -> str:
        return f"<FD c{self.cluster} r{self.resource}>"

    @property
    def cluster(self) -> int:
        """Return the cluster part of the address."""
        cluster, _ = self.parse(self.value)
        return cluster

    @property
    def resource(self) -> int:
        """Return the resource part of the address."""
        _, resource = self.parse(self.value)
        return resource

    @staticmethod
    def parse(address: int) -> tuple[int, int]:
        """Parse the FD address and return the cluster and resource."""
        cluster = (address >> 6) & 0xFFFF
        resource = (address >> 22) & 0x3FF
        return cluster, resource

    @classmethod
    def make(cls, cluster: int, resource: int) -> int:
        """Create a FD address from a cluster and resource."""
        return cls.__type__ | ((cluster & 0xFFFF) << 6) | ((resource & 0x3FF) << 22)


class PointerBlock2Addr(PointerBlockAddr):
    """PB2 address, 32-bit variant. Only used in VMFS5.

    PB (pointer block) are used for indirect addressing of file blocks. They can point to other PBs or file blocks.

    Encoding for PB and PB2 is the same, but the address type differs.
    """

    __type__ = FS3_AddrType.POINTER2_BLOCK

    def __repr__(self) -> str:
        return f"<PB2 cow={self.cow} c{self.cluster} r{self.resource}>"


class PointerBlock2Addr64(PointerBlockAddr64):
    """PB2 address, 64-bit variant. Only used in VMFS6.

    PB (pointer block) are used for indirect addressing of file blocks. They can point to other PBs or file blocks.

    Encoding for PB and PB2 is the same, but the address type differs.
    """

    __type__ = FS3_AddrType.POINTER2_BLOCK

    def __repr__(self) -> str:
        return f"<PB2 cow={self.cow} c{self.cluster} r{self.resource}>"


class JournalBlockAddr(Address):
    """JB address. Only used in VMFS6.

    .. rubric :: Encoding
    .. code-block:: c

        struct FS3_JournalBlockAddr {
            uint32 addrType     : 3;
            uint32 jbClusterNum : 13;
            uint32 unused       : 10;
            uint32 jbNum        : 6;
        };

        0b00000000 00000000 00000000 00000111  (type)
        0b00000000 00000000 11111111 11111000  (cluster)
        0b00000011 11111111 00000000 00000000  (unused/unknown)
        0b11111100 00000000 00000000 00000000  (resource)
    """

    __type__ = FS3_AddrType.JOURNAL_BLOCK

    def __repr__(self) -> str:
        return f"<JB c{self.cluster} r{self.resource}>"

    @property
    def cluster(self) -> int:
        """Return the cluster part of the address."""
        cluster, _ = self.parse(self.value)
        return cluster

    @property
    def resource(self) -> int:
        """Return the resource part of the address."""
        _, resource = self.parse(self.value)
        return resource

    @staticmethod
    def parse(address: int) -> tuple[int, int]:
        """Parse the JB address and return the cluster and resource."""
        cluster = (address >> 3) & 0x1FFF
        resource = (address >> 26) & 0x3F
        return cluster, resource

    @classmethod
    def make(cls, cluster: int, resource: int) -> int:
        """Create a JB address from a cluster and resource."""
        return cls.__type__ | ((cluster & 0x1FFF) << 3) | ((resource & 0x3F) << 26)


class LargeFileBlockAddr(Address):
    """LFB address. Only used in VMFS6.

    LFB (large file block) addresses are used to reference file blocks larger than the standard file block size.
    They are used for large files, such as virtual disks.

    .. rubric :: Encoding
    .. code-block:: c

        struct FS3_LargeFileBlockAddr64 {
            uint64 addrType     : 3;
            uint64 unused1      : 2;
            uint64 copyOnWrite  : 1;
            uint64 unused2      : 1;
            uint64 toBeZeroed   : 8;
            uint64 blockNum     : 31;
            uint64 unused3      : 18;
        };

        0b00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000111  (type)
        0b00000000 00000000 00000000 00000000 00000000 00000000 00000000 00011000  (unused/unknown)
        0b00000000 00000000 00000000 00000000 00000000 00000000 00000000 00100000  (cow)
        0b00000000 00000000 00000000 00000000 00000000 00000000 00000000 01000000  (unused/unknown)
        0b00000000 00000000 00000000 00000000 00000000 00000000 01111111 10000000  (tbz bitmap)
        0b00000000 00000000 00111111 11111111 11111111 11111111 10000000 00000000  (block)
        0b11111111 11111111 11000000 00000000 00000000 00000000 00000000 00000000  (unused/unknown)
    """

    __type__ = FS3_AddrType.LARGE_FILE_BLOCK

    def __repr__(self) -> str:
        return f"<LFB tbz={self.tbz:#x} cow={self.cow} {self.block}>"

    @property
    def cow(self) -> int:
        """Parse the LFB address and return the copy-on-write value."""
        return (self.value >> 5) & 0x1

    @property
    def tbz(self) -> int:
        """Parse the LFB address and return the to-be-zeroed value."""
        return (self.value >> 7) & 0xFF

    @property
    def block(self) -> int:
        return self.parse(self.value)

    @staticmethod
    def parse(address: int) -> int:
        """Parse the LFB address and return the block."""
        return (address >> 15) & 0x7FFFFFFF

    @classmethod
    def make(cls, block: int, cow: int = 0, tbz: int = 0) -> int:
        """Create a LFB address from a block."""
        return cls.__type__ | ((cow & 0x1) << 5) | ((tbz & 0xFF) << 7) | ((block & 0x7FFFFFFF) << 15)



# Generated by cstruct-stubgen
from typing import BinaryIO, Literal, TypeAlias, overload

import dissect.cstruct as __cs__

class _c_vmfs(__cs__.cstruct):
    rootDirDescAddr: Literal[4] = ...
    fbbDescAddr: Literal[4194308] = ...
    fdbcDescAddr: Literal[8388612] = ...
    pbcDescAddr: Literal[12582916] = ...
    sbDescAddr: Literal[16777220] = ...
    vhDescAddr: Literal[20971524] = ...
    pb2DescAddr: Literal[25165828] = ...
    sdDirDescAddr: Literal[29360132] = ...
    jbDescAddr: Literal[33554436] = ...
    unmapDescAddr: Literal[37748740] = ...
    dfDirDescAddr: Literal[41943044] = ...
    FS3_FS_HEADER_OFFSET: Literal[2097152] = ...
    FS3_VMFS6_MAJOR_VERSION: Literal[24] = ...
    VMFS_MAGIC_NUMBER: Literal[799797598] = ...
    VMFSL_MAGIC_NUMBER: Literal[799797599] = ...
    VMFS6_MAGIC_NUMBER: Literal[799797600] = ...
    VMFS6L_MAGIC_NUMBER: Literal[799797601] = ...
    FS3_RFMD_SIGNATURE: Literal[1919315300] = ...
    FS3_RCMD_SIGNATURE: Literal[1919118692] = ...
    FS3_FDMD_SIGNATURE: Literal[1717857636] = ...
    FS3_MAX_FILE_NAME_LENGTH: Literal[128] = ...
    FS6_MAX_FILE_NAME_LENGTH: Literal[256] = ...
    FS6_DIR_HEADER_VERSION: Literal[16056321] = ...
    FS6_DIR_HEADER_DEBUG_VERSION: Literal[16629761] = ...
    FS6_DIR_HEADER_BLOCK_SIZE: Literal[65536] = ...
    FS6_DIR_HASH_MAX_ENTRIES: Literal[16001] = ...
    FS6_DIR_HASH_ROOT_RESERVED: Literal[28] = ...
    FS6_DIR_HASH_MAX_ROOT_ENTRIES: Literal[15973] = ...
    FS6_DIR_LINKS_PER_GROUP: Literal[12] = ...
    class UUID(__cs__.Structure):
        timeLo: _c_vmfs.uint32
        timeHi: _c_vmfs.uint32
        rand: _c_vmfs.uint16
        macAddr: __cs__.CharArray
        @overload
        def __init__(
            self,
            timeLo: _c_vmfs.uint32 | None = ...,
            timeHi: _c_vmfs.uint32 | None = ...,
            rand: _c_vmfs.uint16 | None = ...,
            macAddr: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class SCSI_DiskId(__cs__.Structure):
        type: _c_vmfs.uint8
        len: _c_vmfs.uint8
        lun: _c_vmfs.uint16
        deviceType: _c_vmfs.uint8
        scsiLevel: _c_vmfs.uint8
        vendor: __cs__.CharArray
        model: __cs__.CharArray
        revision: __cs__.CharArray
        id: __cs__.CharArray
        @overload
        def __init__(
            self,
            type: _c_vmfs.uint8 | None = ...,
            len: _c_vmfs.uint8 | None = ...,
            lun: _c_vmfs.uint16 | None = ...,
            deviceType: _c_vmfs.uint8 | None = ...,
            scsiLevel: _c_vmfs.uint8 | None = ...,
            vendor: __cs__.CharArray | None = ...,
            model: __cs__.CharArray | None = ...,
            revision: __cs__.CharArray | None = ...,
            id: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    FS3_Address: TypeAlias = _c_vmfs.uint32
    FS3_FileBlockAddr: TypeAlias = _c_vmfs.uint32
    FS3_FDAddr: TypeAlias = _c_vmfs.uint32
    class FS3_VolAddress(__cs__.Structure):
        offset: _c_vmfs.uint64
        @overload
        def __init__(self, offset: _c_vmfs.uint64 | None = ...): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class FS3_DescriptorType(__cs__.Enum):
        INVALID = ...
        VOLUME = ...
        DIRECTORY = ...
        REGFILE = ...
        SYMLINK = ...
        SYSFILE = ...
        RDM = ...
        SENTINEL = ...

    class FS3_AddrType(__cs__.Enum):
        INVALID = ...
        FILE_BLOCK = ...
        SMALL_FILE_BLOCK = ...
        SUB_BLOCK = ...
        POINTER_BLOCK = ...
        FILE_DESCRIPTOR = ...
        POINTER2_BLOCK = ...
        JOURNAL_BLOCK = ...
        LARGE_FILE_BLOCK = ...
        SENTINEL = ...

    class FS3_ZeroLevelAddrType(__cs__.Enum):
        INVALID = ...
        FILE_BLOCK = ...
        SUB_BLOCK = ...
        POINTER_BLOCK = ...
        POINTER2_BLOCK = ...
        POINTER_BLOCK_DOUBLE = ...
        FILE_DESCRIPTOR_RESIDENT = ...

    class FS3_Config(__cs__.Flag):
        UNKNOWN1 = ...
        UNKNOWN2 = ...
        DENSE_SBPC = ...
        MAINTENANCE = ...
        UNKNOWN10 = ...
        UNKNOWN20 = ...
        UNKNOWN40 = ...
        UNKNOWN80 = ...
        LOCAL = ...
        UNKNOWN200 = ...
        SHARED = ...
        OSDATA = ...
        VSAN = ...
        UNKNOWN2000 = ...
        UNKNOWN4000 = ...

    class FDS_VolInfo(__cs__.Structure):
        id: __cs__.CharArray
        @overload
        def __init__(self, id: __cs__.CharArray | None = ...): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class FS3_Checksum(__cs__.Structure):
        value: _c_vmfs.uint64
        checksumGen: _c_vmfs.uint64
        @overload
        def __init__(self, value: _c_vmfs.uint64 | None = ..., checksumGen: _c_vmfs.uint64 | None = ...): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class FS3_UnmapPriority(__cs__.Enum):
        NONE = ...
        LOW = ...
        MEDIUM = ...
        HIGH = ...

    class FS3_UnmapMethod(__cs__.Enum):
        PRIORITY = ...
        FIXED = ...

    class FS3_Descriptor(__cs__.Structure):
        magic: _c_vmfs.uint32
        majorVersion: _c_vmfs.uint32
        minorVersion: _c_vmfs.uint8
        uuid: _c_vmfs.UUID
        config: _c_vmfs.FS3_Config
        fsLabel: __cs__.CharArray
        diskBlockSize: _c_vmfs.uint32
        fileBlockSize: _c_vmfs.uint64
        creationTime: _c_vmfs.uint32
        snapID: _c_vmfs.uint32
        volInfo: _c_vmfs.FDS_VolInfo
        fdcClusterGroupOffset: _c_vmfs.uint32
        fdcClustersPerGroup: _c_vmfs.uint32
        subBlockSize: _c_vmfs.uint32
        maxJournalSlotsPerTxn: _c_vmfs.uint32
        pb2VolAddr: _c_vmfs.uint64
        pb2FDAddr: _c_vmfs.uint32
        hostUuid: __cs__.CharArray
        gblGeneration: _c_vmfs.uint64
        sddVolAddr: _c_vmfs.uint64
        sddFDAddr: _c_vmfs.uint32
        checksumType: _c_vmfs.uint8
        unmapPriority: _c_vmfs.FS3_UnmapPriority
        pad1: __cs__.CharArray
        checksumGen: _c_vmfs.uint64
        checksum: _c_vmfs.FS3_Checksum
        physDiskBlockSize: _c_vmfs.uint32
        mdAlignment: _c_vmfs.uint32
        sfbToLfbShift: _c_vmfs.uint16
        reserved16_1: _c_vmfs.uint16
        reserved16_2: _c_vmfs.uint16
        ptrBlockShift: _c_vmfs.uint16
        sfbAddrBits: _c_vmfs.uint16
        reserved16_3: _c_vmfs.uint16
        tbzGranularity: _c_vmfs.uint32
        journalBlockSize: _c_vmfs.uint32
        leaseIntervalMs: _c_vmfs.uint32
        reclaimWindowMs: _c_vmfs.uint32
        localStampUS: _c_vmfs.uint64
        localMountOwnerMacAddr: __cs__.CharArray
        unmapMethod: _c_vmfs.FS3_UnmapMethod
        unmapBandwidth: _c_vmfs.uint32
        unmapBandwidthMin: _c_vmfs.uint32
        unmapBandwidthMax: _c_vmfs.uint32
        @overload
        def __init__(
            self,
            magic: _c_vmfs.uint32 | None = ...,
            majorVersion: _c_vmfs.uint32 | None = ...,
            minorVersion: _c_vmfs.uint8 | None = ...,
            uuid: _c_vmfs.UUID | None = ...,
            config: _c_vmfs.FS3_Config | None = ...,
            fsLabel: __cs__.CharArray | None = ...,
            diskBlockSize: _c_vmfs.uint32 | None = ...,
            fileBlockSize: _c_vmfs.uint64 | None = ...,
            creationTime: _c_vmfs.uint32 | None = ...,
            snapID: _c_vmfs.uint32 | None = ...,
            volInfo: _c_vmfs.FDS_VolInfo | None = ...,
            fdcClusterGroupOffset: _c_vmfs.uint32 | None = ...,
            fdcClustersPerGroup: _c_vmfs.uint32 | None = ...,
            subBlockSize: _c_vmfs.uint32 | None = ...,
            maxJournalSlotsPerTxn: _c_vmfs.uint32 | None = ...,
            pb2VolAddr: _c_vmfs.uint64 | None = ...,
            pb2FDAddr: _c_vmfs.uint32 | None = ...,
            hostUuid: __cs__.CharArray | None = ...,
            gblGeneration: _c_vmfs.uint64 | None = ...,
            sddVolAddr: _c_vmfs.uint64 | None = ...,
            sddFDAddr: _c_vmfs.uint32 | None = ...,
            checksumType: _c_vmfs.uint8 | None = ...,
            unmapPriority: _c_vmfs.FS3_UnmapPriority | None = ...,
            pad1: __cs__.CharArray | None = ...,
            checksumGen: _c_vmfs.uint64 | None = ...,
            checksum: _c_vmfs.FS3_Checksum | None = ...,
            physDiskBlockSize: _c_vmfs.uint32 | None = ...,
            mdAlignment: _c_vmfs.uint32 | None = ...,
            sfbToLfbShift: _c_vmfs.uint16 | None = ...,
            reserved16_1: _c_vmfs.uint16 | None = ...,
            reserved16_2: _c_vmfs.uint16 | None = ...,
            ptrBlockShift: _c_vmfs.uint16 | None = ...,
            sfbAddrBits: _c_vmfs.uint16 | None = ...,
            reserved16_3: _c_vmfs.uint16 | None = ...,
            tbzGranularity: _c_vmfs.uint32 | None = ...,
            journalBlockSize: _c_vmfs.uint32 | None = ...,
            leaseIntervalMs: _c_vmfs.uint32 | None = ...,
            reclaimWindowMs: _c_vmfs.uint32 | None = ...,
            localStampUS: _c_vmfs.uint64 | None = ...,
            localMountOwnerMacAddr: __cs__.CharArray | None = ...,
            unmapMethod: _c_vmfs.FS3_UnmapMethod | None = ...,
            unmapBandwidth: _c_vmfs.uint32 | None = ...,
            unmapBandwidthMin: _c_vmfs.uint32 | None = ...,
            unmapBandwidthMax: _c_vmfs.uint32 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class FS3_ResourceTypeID(__cs__.Enum):
        INVALID = ...
        FILE_BLOCK = ...
        SMALL_FILE_BLOCK = ...
        SUB_BLOCK = ...
        PTR_BLOCK = ...
        FILE_DESC = ...
        PTR2_BLOCK = ...
        JOURNAL_BLOCK = ...
        LARGE_FILE_BLOCK = ...
        SENTINEL = ...

    class Res3PriorityType(__cs__.Enum):
        DEFAULT = ...
        JOURNAL = ...
        MAX = ...

    class FS3_ResFileMetadata(__cs__.Structure):
        resourcesPerCluster: _c_vmfs.uint32
        clustersPerGroup: _c_vmfs.uint32
        clusterGroupOffset: _c_vmfs.uint32
        resourceSize: _c_vmfs.uint32
        clusterGroupSize: _c_vmfs.uint32
        numResourcesLo: _c_vmfs.uint32
        numClusterGroups: _c_vmfs.uint32
        numResourcesHi: _c_vmfs.uint32
        signature: _c_vmfs.uint32
        version: _c_vmfs.uint32
        flags: _c_vmfs.uint32
        affinityPerCluster: _c_vmfs.uint16
        affinityPerResource: _c_vmfs.uint16
        bitsPerResource: _c_vmfs.uint32
        childMetaOffset: _c_vmfs.uint32
        parentResourcesPerCluster: _c_vmfs.uint32
        parentClustersPerGroup: _c_vmfs.uint32
        parentClusterGroupSize: _c_vmfs.uint32
        _unknown: __cs__.CharArray
        convertedClusters: _c_vmfs.uint32
        @overload
        def __init__(
            self,
            resourcesPerCluster: _c_vmfs.uint32 | None = ...,
            clustersPerGroup: _c_vmfs.uint32 | None = ...,
            clusterGroupOffset: _c_vmfs.uint32 | None = ...,
            resourceSize: _c_vmfs.uint32 | None = ...,
            clusterGroupSize: _c_vmfs.uint32 | None = ...,
            numResourcesLo: _c_vmfs.uint32 | None = ...,
            numClusterGroups: _c_vmfs.uint32 | None = ...,
            numResourcesHi: _c_vmfs.uint32 | None = ...,
            signature: _c_vmfs.uint32 | None = ...,
            version: _c_vmfs.uint32 | None = ...,
            flags: _c_vmfs.uint32 | None = ...,
            affinityPerCluster: _c_vmfs.uint16 | None = ...,
            affinityPerResource: _c_vmfs.uint16 | None = ...,
            bitsPerResource: _c_vmfs.uint32 | None = ...,
            childMetaOffset: _c_vmfs.uint32 | None = ...,
            parentResourcesPerCluster: _c_vmfs.uint32 | None = ...,
            parentClustersPerGroup: _c_vmfs.uint32 | None = ...,
            parentClusterGroupSize: _c_vmfs.uint32 | None = ...,
            _unknown: __cs__.CharArray | None = ...,
            convertedClusters: _c_vmfs.uint32 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class FS3_ResourceClusterMD(__cs__.Structure):
        clusterNum: _c_vmfs.uint32
        totalResources: _c_vmfs.uint32
        freeResources: _c_vmfs.uint32
        nextFreeIdx: _c_vmfs.uint32
        bitmap: __cs__.Array[_c_vmfs.uint8]
        typeData: __cs__.Array[_c_vmfs.uint8]
        priority: _c_vmfs.Res3PriorityType
        @overload
        def __init__(
            self,
            clusterNum: _c_vmfs.uint32 | None = ...,
            totalResources: _c_vmfs.uint32 | None = ...,
            freeResources: _c_vmfs.uint32 | None = ...,
            nextFreeIdx: _c_vmfs.uint32 | None = ...,
            bitmap: __cs__.Array[_c_vmfs.uint8] | None = ...,
            typeData: __cs__.Array[_c_vmfs.uint8] | None = ...,
            priority: _c_vmfs.Res3PriorityType | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class Res3_ClusterMetaVMFS6(__cs__.Structure):
        _pad: __cs__.CharArray
        signature: _c_vmfs.uint32
        priority: _c_vmfs.uint32
        clusterNum: _c_vmfs.uint64
        totalResources: _c_vmfs.uint16
        freeResources: _c_vmfs.uint16
        unk1: _c_vmfs.uint16
        affinityCount: _c_vmfs.uint16
        _unk: __cs__.CharArray
        @overload
        def __init__(
            self,
            _pad: __cs__.CharArray | None = ...,
            signature: _c_vmfs.uint32 | None = ...,
            priority: _c_vmfs.uint32 | None = ...,
            clusterNum: _c_vmfs.uint64 | None = ...,
            totalResources: _c_vmfs.uint16 | None = ...,
            freeResources: _c_vmfs.uint16 | None = ...,
            unk1: _c_vmfs.uint16 | None = ...,
            affinityCount: _c_vmfs.uint16 | None = ...,
            _unk: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class FS3_HBGen(__cs__.Structure):
        gen: _c_vmfs.uint64
        @overload
        def __init__(self, gen: _c_vmfs.uint64 | None = ...): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class FS3_LockHolder(__cs__.Structure):
        hbAddr: _c_vmfs.FS3_VolAddress
        hbGen: _c_vmfs.FS3_HBGen
        uid: _c_vmfs.UUID
        @overload
        def __init__(
            self,
            hbAddr: _c_vmfs.FS3_VolAddress | None = ...,
            hbGen: _c_vmfs.FS3_HBGen | None = ...,
            uid: _c_vmfs.UUID | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class FS3_DiskLock(__cs__.Structure):
        type: _c_vmfs.uint32
        addr: _c_vmfs.FS3_VolAddress
        hbAddr: _c_vmfs.FS3_VolAddress
        hbGen: _c_vmfs.FS3_HBGen
        token: _c_vmfs.uint64
        mode: _c_vmfs.uint32
        owner: _c_vmfs.UUID
        mtime: _c_vmfs.uint64
        numHolders: _c_vmfs.uint32
        class FS3_LockHolder(__cs__.Structure):
            hbAddr: _c_vmfs.FS3_VolAddress
            hbGen: _c_vmfs.FS3_HBGen
            uid: _c_vmfs.UUID
            @overload
            def __init__(
                self,
                hbAddr: _c_vmfs.FS3_VolAddress | None = ...,
                hbGen: _c_vmfs.FS3_HBGen | None = ...,
                uid: _c_vmfs.UUID | None = ...,
            ): ...
            @overload
            def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

        holders: __cs__.Array[FS3_LockHolder]
        _unk: __cs__.CharArray
        gblNumHolders: _c_vmfs.uint32
        gblGen: _c_vmfs.uint64
        gblBrk: _c_vmfs.uint32
        @overload
        def __init__(
            self,
            type: _c_vmfs.uint32 | None = ...,
            addr: _c_vmfs.FS3_VolAddress | None = ...,
            hbAddr: _c_vmfs.FS3_VolAddress | None = ...,
            hbGen: _c_vmfs.FS3_HBGen | None = ...,
            token: _c_vmfs.uint64 | None = ...,
            mode: _c_vmfs.uint32 | None = ...,
            owner: _c_vmfs.UUID | None = ...,
            mtime: _c_vmfs.uint64 | None = ...,
            numHolders: _c_vmfs.uint32 | None = ...,
            holders: __cs__.Array[FS3_LockHolder] | None = ...,
            _unk: __cs__.CharArray | None = ...,
            gblNumHolders: _c_vmfs.uint32 | None = ...,
            gblGen: _c_vmfs.uint64 | None = ...,
            gblBrk: _c_vmfs.uint32 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class FS3_Heartbeat(__cs__.Structure):
        state: _c_vmfs.uint32
        addr: _c_vmfs.FS3_VolAddress
        hbGen: _c_vmfs.FS3_HBGen
        stamp: _c_vmfs.uint64
        owner: _c_vmfs.UUID
        journalAddr: _c_vmfs.uint32
        driveX: _c_vmfs.uint32
        driveY: _c_vmfs.uint8
        lockImpl: _c_vmfs.uint8
        ip: __cs__.CharArray
        journalVolAddr: _c_vmfs.uint64
        replayHost: _c_vmfs.UUID
        _unk: __cs__.CharArray
        replayHostHB: _c_vmfs.uint64
        replayHostHBgen: _c_vmfs.FS3_HBGen
        @overload
        def __init__(
            self,
            state: _c_vmfs.uint32 | None = ...,
            addr: _c_vmfs.FS3_VolAddress | None = ...,
            hbGen: _c_vmfs.FS3_HBGen | None = ...,
            stamp: _c_vmfs.uint64 | None = ...,
            owner: _c_vmfs.UUID | None = ...,
            journalAddr: _c_vmfs.uint32 | None = ...,
            driveX: _c_vmfs.uint32 | None = ...,
            driveY: _c_vmfs.uint8 | None = ...,
            lockImpl: _c_vmfs.uint8 | None = ...,
            ip: __cs__.CharArray | None = ...,
            journalVolAddr: _c_vmfs.uint64 | None = ...,
            replayHost: _c_vmfs.UUID | None = ...,
            _unk: __cs__.CharArray | None = ...,
            replayHostHB: _c_vmfs.uint64 | None = ...,
            replayHostHBgen: _c_vmfs.FS3_HBGen | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class FS3_FileMetadata(__cs__.Structure):
        descAddr: _c_vmfs.uint32
        generation: _c_vmfs.uint32
        linkCount: _c_vmfs.uint32
        type: _c_vmfs.FS3_DescriptorType
        flags: _c_vmfs.uint32
        fileLength: _c_vmfs.uint64
        blockSize: _c_vmfs.uint64
        numBlocks: _c_vmfs.uint64
        mtime: _c_vmfs.uint32
        ctime: _c_vmfs.uint32
        atime: _c_vmfs.uint32
        uid: _c_vmfs.uint32
        gid: _c_vmfs.uint32
        mode: _c_vmfs.uint32
        zeroLevelAddrType: _c_vmfs.uint32
        numTBZBlocksLo: _c_vmfs.uint32
        numCOWBlocksLo: _c_vmfs.uint32
        newSinceEpochLo: _c_vmfs.uint32
        numTBZBlocksHi: _c_vmfs.uint32
        numCOWBlocksHi: _c_vmfs.uint32
        numPointerBlocks: _c_vmfs.uint32
        newSinceEpochHi: _c_vmfs.uint32
        signature: _c_vmfs.uint32
        affinityFD: _c_vmfs.uint32
        tbzGranularityShift: _c_vmfs.uint32
        parentFD: _c_vmfs.uint32
        lastSFBClusterNum: _c_vmfs.uint32
        _unk4: _c_vmfs.uint32
        _unk5: _c_vmfs.uint32
        affinityBitmap: _c_vmfs.uint32
        numPreAllocBlocks: _c_vmfs.uint8
        _unk7: _c_vmfs.uint8
        _unk8: _c_vmfs.uint8
        _unk9: _c_vmfs.uint8
        _unk10: _c_vmfs.uint8
        blockOffsetShift: _c_vmfs.uint8
        numTracked: _c_vmfs.uint8
        _unk12: _c_vmfs.uint8
        numLFB: _c_vmfs.uint32
        _unk13: __cs__.CharArray
        lastFreeSFBC: _c_vmfs.uint32
        @overload
        def __init__(
            self,
            descAddr: _c_vmfs.uint32 | None = ...,
            generation: _c_vmfs.uint32 | None = ...,
            linkCount: _c_vmfs.uint32 | None = ...,
            type: _c_vmfs.FS3_DescriptorType | None = ...,
            flags: _c_vmfs.uint32 | None = ...,
            fileLength: _c_vmfs.uint64 | None = ...,
            blockSize: _c_vmfs.uint64 | None = ...,
            numBlocks: _c_vmfs.uint64 | None = ...,
            mtime: _c_vmfs.uint32 | None = ...,
            ctime: _c_vmfs.uint32 | None = ...,
            atime: _c_vmfs.uint32 | None = ...,
            uid: _c_vmfs.uint32 | None = ...,
            gid: _c_vmfs.uint32 | None = ...,
            mode: _c_vmfs.uint32 | None = ...,
            zeroLevelAddrType: _c_vmfs.uint32 | None = ...,
            numTBZBlocksLo: _c_vmfs.uint32 | None = ...,
            numCOWBlocksLo: _c_vmfs.uint32 | None = ...,
            newSinceEpochLo: _c_vmfs.uint32 | None = ...,
            numTBZBlocksHi: _c_vmfs.uint32 | None = ...,
            numCOWBlocksHi: _c_vmfs.uint32 | None = ...,
            numPointerBlocks: _c_vmfs.uint32 | None = ...,
            newSinceEpochHi: _c_vmfs.uint32 | None = ...,
            signature: _c_vmfs.uint32 | None = ...,
            affinityFD: _c_vmfs.uint32 | None = ...,
            tbzGranularityShift: _c_vmfs.uint32 | None = ...,
            parentFD: _c_vmfs.uint32 | None = ...,
            lastSFBClusterNum: _c_vmfs.uint32 | None = ...,
            _unk4: _c_vmfs.uint32 | None = ...,
            _unk5: _c_vmfs.uint32 | None = ...,
            affinityBitmap: _c_vmfs.uint32 | None = ...,
            numPreAllocBlocks: _c_vmfs.uint8 | None = ...,
            _unk7: _c_vmfs.uint8 | None = ...,
            _unk8: _c_vmfs.uint8 | None = ...,
            _unk9: _c_vmfs.uint8 | None = ...,
            _unk10: _c_vmfs.uint8 | None = ...,
            blockOffsetShift: _c_vmfs.uint8 | None = ...,
            numTracked: _c_vmfs.uint8 | None = ...,
            _unk12: _c_vmfs.uint8 | None = ...,
            numLFB: _c_vmfs.uint32 | None = ...,
            _unk13: __cs__.CharArray | None = ...,
            lastFreeSFBC: _c_vmfs.uint32 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class FS3_RawDiskMap(__cs__.Structure):
        diskId: _c_vmfs.SCSI_DiskId
        @overload
        def __init__(self, diskId: _c_vmfs.SCSI_DiskId | None = ...): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class FS3_DirEntry(__cs__.Structure):
        type: _c_vmfs.FS3_DescriptorType
        descAddr: _c_vmfs.uint32
        generation: _c_vmfs.uint32
        name: __cs__.CharArray
        @overload
        def __init__(
            self,
            type: _c_vmfs.FS3_DescriptorType | None = ...,
            descAddr: _c_vmfs.uint32 | None = ...,
            generation: _c_vmfs.uint32 | None = ...,
            name: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class FS6_DirBlockType(__cs__.Enum):
        DIRENT = ...
        LINK = ...
        ALLOCATION_MAP = ...

    class FS6_DirEntry(__cs__.Structure):
        type: _c_vmfs.FS3_DescriptorType
        address: _c_vmfs.uint32
        generation: _c_vmfs.uint32
        hashIndex: _c_vmfs.uint16
        linkHash: _c_vmfs.uint16
        offset: _c_vmfs.uint64
        name: __cs__.CharArray
        _unk1: _c_vmfs.uint64
        @overload
        def __init__(
            self,
            type: _c_vmfs.FS3_DescriptorType | None = ...,
            address: _c_vmfs.uint32 | None = ...,
            generation: _c_vmfs.uint32 | None = ...,
            hashIndex: _c_vmfs.uint16 | None = ...,
            linkHash: _c_vmfs.uint16 | None = ...,
            offset: _c_vmfs.uint64 | None = ...,
            name: __cs__.CharArray | None = ...,
            _unk1: _c_vmfs.uint64 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class FS6_DirHeader(__cs__.Structure):
        version: _c_vmfs.uint32
        numEntries: _c_vmfs.uint32
        numAllocated: _c_vmfs.uint32
        numAllocationMapBlocks: _c_vmfs.uint32
        allocationMapBlocks: __cs__.Array[_c_vmfs.uint32]
        _gap0: __cs__.CharArray
        checksum: _c_vmfs.uint64
        checksumGen: _c_vmfs.uint64
        selfEntry: _c_vmfs.FS6_DirEntry
        parentEntry: _c_vmfs.FS6_DirEntry
        @overload
        def __init__(
            self,
            version: _c_vmfs.uint32 | None = ...,
            numEntries: _c_vmfs.uint32 | None = ...,
            numAllocated: _c_vmfs.uint32 | None = ...,
            numAllocationMapBlocks: _c_vmfs.uint32 | None = ...,
            allocationMapBlocks: __cs__.Array[_c_vmfs.uint32] | None = ...,
            _gap0: __cs__.CharArray | None = ...,
            checksum: _c_vmfs.uint64 | None = ...,
            checksumGen: _c_vmfs.uint64 | None = ...,
            selfEntry: _c_vmfs.FS6_DirEntry | None = ...,
            parentEntry: _c_vmfs.FS6_DirEntry | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class FS6_DirBlockHeader(__cs__.Structure):
        version: _c_vmfs.uint16
        type: _c_vmfs.uint16
        totalSlots: _c_vmfs.uint16
        freeSlots: _c_vmfs.uint16
        bitmap: __cs__.CharArray
        @overload
        def __init__(
            self,
            version: _c_vmfs.uint16 | None = ...,
            type: _c_vmfs.uint16 | None = ...,
            totalSlots: _c_vmfs.uint16 | None = ...,
            freeSlots: _c_vmfs.uint16 | None = ...,
            bitmap: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class FS6_DirLink(__cs__.Structure):
        location: _c_vmfs.uint32
        hash: _c_vmfs.uint16
        @overload
        def __init__(self, location: _c_vmfs.uint32 | None = ..., hash: _c_vmfs.uint16 | None = ...): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class FS6_DirLinkGroup(__cs__.Structure):
        hashIndex: _c_vmfs.uint16
        totalLinks: _c_vmfs.uint8
        freeLinks: _c_vmfs.uint8
        nextFreeIdx: _c_vmfs.uint8
        pad1: _c_vmfs.uint8
        class FS6_DirLink(__cs__.Structure):
            location: _c_vmfs.uint32
            hash: _c_vmfs.uint16
            @overload
            def __init__(self, location: _c_vmfs.uint32 | None = ..., hash: _c_vmfs.uint16 | None = ...): ...
            @overload
            def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

        links: __cs__.Array[FS6_DirLink]
        nextGroup: _c_vmfs.uint32
        pad2: _c_vmfs.uint16
        @overload
        def __init__(
            self,
            hashIndex: _c_vmfs.uint16 | None = ...,
            totalLinks: _c_vmfs.uint8 | None = ...,
            freeLinks: _c_vmfs.uint8 | None = ...,
            nextFreeIdx: _c_vmfs.uint8 | None = ...,
            pad1: _c_vmfs.uint8 | None = ...,
            links: __cs__.Array[FS6_DirLink] | None = ...,
            nextGroup: _c_vmfs.uint32 | None = ...,
            pad2: _c_vmfs.uint16 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

# Technically `c_vmfs` is an instance of `_c_vmfs`, but then we can't use it in type hints
c_vmfs: TypeAlias = _c_vmfs

FS3_Config = c_vmfs.FS3_Config
FS3_AddrType = c_vmfs.FS3_AddrType
FS3_ResourceTypeID = c_vmfs.FS3_ResourceTypeID
FS3_ZeroLevelAddrType = c_vmfs.FS3_ZeroLevelAddrType
FS3_DescriptorType = c_vmfs.FS3_DescriptorType
FS6_DirBlockType = c_vmfs.FS6_DirBlockType
